%module senna

%header %{

#include "senna.h"
typedef int bool;
typedef sen_index Index;
typedef sen_records Records;
typedef sen_values Values;
typedef sen_sym Sym;
typedef sen_set Set;
typedef struct {
  const void *key;
  int ikey;
  int size;
} sen_key;

typedef struct {
  Records *records;
  const sen_recordh *rh;
} Record;

typedef struct {
  Set *set;
  sen_set_cursor *cursor;
} SetCursor;

#define RC_HANDLER(rc) { \
  switch (rc) { \
  case sen_success : \
    break; \
  case sen_memory_exhausted : \
    SWIG_exception_(SWIG_MemoryError, "memory exhausted"); \
    break; \
  case sen_invalid_format : \
    SWIG_exception_(SWIG_TypeError, "invalid format"); \
    break; \
  case sen_file_operation_error : \
    SWIG_exception_(SWIG_IOError, "file operation error"); \
    break; \
  case sen_invalid_argument : \
    SWIG_exception_(SWIG_ValueError, "invalid argument"); \
    break; \
  default : \
    SWIG_exception_(SWIG_UnknownError, "other error"); \
    break; \
  } \
}

#define VALIDATE_KEY_SIZE(key,key_size) { \
  SWIG_contract_assert(key_size == key.size \
                       || (!key.size && strlen(key.key) >= key_size), \
                       "key size unmatch"); \
}

static int
sample_group_func(sen_records *r, const sen_recordh *rh, void **key, void *arg)
{
  sen_recinfo ri;
  if (sen_record_info(r, rh, &ri)) { return 1; }
  memcpy(key, &ri.n_subrecs, sizeof(int));
  return 0;
}

static int
sample_sort_func(sen_records *r, sen_recordh *a, sen_recordh *b, void *arg)
{
  sen_recinfo ria, rib;
  if (sen_record_info(r, a, &ria)) { return -1; }
  if (sen_record_info(r, b, &rib)) { return 1; }
  if (ria.n_subrecs == rib.n_subrecs) {
    return ria.score - rib.score;
  } else {
    return ria.n_subrecs - rib.n_subrecs;
  }
}

static int
sample_select_func(sen_records *r, const void *key, int section, void *arg)
{
  int *value;
  if (section >= (sizeof(int) * 8)) { return 0; }
  if (!sen_set_at((sen_set *)arg, key, (void **)&value)) { return 0; }
  return (*value & (1 << section)) ? 1 : 0;
}

%}

%init %{
sen_init();
%}

%include "exception.i"
%include "typemaps.i"

%typemap(in,numinputs=0) sen_key *OUTPUT(sen_key temp), sen_key &OUTPUT(sen_key temp) "$1 = &temp;";

%typemap(ruby,in) sen_key {
  switch (TYPE($input)) {
  case T_FIXNUM :
    $1.ikey = NUM2INT($input);
    $1.key = &$1.ikey;
    $1.size = sizeof(int);
    break;
  case T_STRING :
    $1.key = StringValuePtr($input);
    $1.size = 0;
    break;
  default :
    SWIG_exception_(SWIG_TypeError, "Expected string or fixnum");
    break;
  }
}
%typemap(ruby,out) sen_key {
  if (!$1.key) {
    $result = Qnil;
  } else {
    switch ($1.size) {
    case 0 :
      $result = rb_str_new2((char *)($1.key));
      break;
    case sizeof(int) :
      $result = INT2NUM(*((int *)($1.key)));
      break;
    default :
      $result = rb_str_new((char *)($1.key), $1.size);
      break;
    }
  }
}
%typemap(ruby,argout, fragment="output_helper") sen_key *OUTPUT, sen_key &OUTPUT {
   VALUE o;
   if (!$1->key) {
     o = Qnil;
   } else {
     switch ($1->size) {
     case 0 :
       o = rb_str_new2((char *)($1->key));
       break;
     case sizeof(int) :
       o = INT2NUM(*((int *)($1->key)));
       break;
     default :
       o = rb_str_new((char *)($1->key), $1->size);
       break;
     }
   }
   $result = output_helper($result, o);
}

typedef unsigned sen_id;

#define INDEX_NORMALIZE     0x0001
#define INDEX_SPLIT_ALPHA   0x0002
#define INDEX_SPLIT_DIGIT   0x0004
#define INDEX_SPLIT_SYMBOL  0x0008
#define INDEX_NGRAM         0x0010

%constant int KEY_STRING = 0;
%constant int KEY_INT    = sizeof(int);

enum sen_rc;
%constant sen_rc SEN_SUCCESS = sen_success;
%constant sen_rc SEN_MEMORY_EXHAUSTED = sen_memory_exhausted;
%constant sen_rc SEN_INVALID_FORMAT = sen_invalid_format;
%constant sen_rc SEN_FILE_OPERATION_ERROR = sen_file_operation_error;
%constant sen_rc SEN_INVALID_ARGUMENT = sen_invalid_argument;
%constant sen_rc SEN_OTHER_ERROR = sen_other_error;

enum sen_encoding;
%constant sen_encoding ENC_DEFAULT = sen_enc_default;
%constant sen_encoding ENC_NONE = sen_enc_none;
%constant sen_encoding ENC_EUC_JP = sen_enc_euc_jp;
%constant sen_encoding ENC_UTF8 = sen_enc_utf8;
%constant sen_encoding ENC_SJIS = sen_enc_sjis;

enum sen_rec_unit;
%constant sen_rec_unit REC_DOCUMENT = sen_rec_document;
%constant sen_rec_unit REC_SECTION = sen_rec_section;
%constant sen_rec_unit REC_POSITION = sen_rec_position;
%constant sen_rec_unit REC_USERDEF = sen_rec_userdef;
%constant sen_rec_unit REC_NONE = sen_rec_none;

enum sen_sel_operator;
%constant sen_sel_operator SEL_OR = sen_sel_or;
%constant sen_sel_operator SEL_AND = sen_sel_and;
%constant sen_sel_operator SEL_BUT = sen_sel_but;
%constant sen_sel_operator SEL_ADJUSST = sen_sel_adjust;

enum sen_sel_mode;
%constant sen_sel_mode SEL_EXACT = sen_sel_exact;
%constant sen_sel_mode SEL_PARTIAL = sen_sel_partial;
%constant sen_sel_mode SEL_UNSPLIT = sen_sel_unsplit;
%constant sen_sel_mode SEL_NEAR = sen_sel_near;
%constant sen_sel_mode SEL_SIMILAR = sen_sel_similar;

enum sen_sort_mode;
%constant sen_sort_mode SORT_DESCENDING = sen_sort_descending;
%constant sen_sort_mode SORT_ASCENDING = sen_sort_ascending;

%apply int *OUTPUT { int *key_size, int *flags,
                     int *initial_n_segments, sen_encoding *encoding };
typedef struct {
  Sym *keys;
  Sym *lexicon;
  %extend {
    Index(const char *path) {
      return sen_index_open(path);
    }
    static Index *open(const char *path) {
      return sen_index_open(path);
    }
    static Index *create(const char *path, int key_size=0, int flags=0,
          int initial_n_segments=0, sen_encoding encoding=0) {
      SWIG_contract_assert(0 <= key_size && key_size < 8192, "invalid key_size");
      return sen_index_create(path, key_size, flags, initial_n_segments, encoding);
    }
    static Index *open_with_keys(const char *path, Sym *keys) {
      return sen_index_open_with_keys(path, keys);
    }
    static Index *create_with_keys(const char *path, Sym *keys, int flags=0,
                                   int initial_n_segments=0, sen_encoding encoding=0) {
      return sen_index_create_with_keys(path, keys, flags, initial_n_segments, encoding);
    }
    void info(int *key_size, int *flags,
		int *initial_n_segments, sen_encoding *encoding) {
      sen_index_info(self, key_size, flags, initial_n_segments, encoding);
    }
    sen_rc upd(sen_key key, const char *oldvalue, const char *newvalue) {
      int key_size;
      sen_index_info(self, &key_size, NULL, NULL, NULL);
      VALIDATE_KEY_SIZE(key,key_size);
      return sen_index_upd(self, key.key, oldvalue, newvalue);
    }
    Records *sel(const char *string) {
      return sen_index_sel(self, string);
    }
    ~Index() {
      sen_index_close(self);
    }
    sen_rc update(sen_key key, unsigned section,
                  Values *oldvalues, Values *newvalues) {
      int key_size;
      sen_index_info(self, &key_size, NULL, NULL, NULL);
      VALIDATE_KEY_SIZE(key,key_size);
      return sen_index_update(self, key.key, section, oldvalues, newvalues);
    }
    sen_rc select(const char *string, Records *r, sen_sel_operator op,
                  sen_sel_mode mode=sen_sel_exact,
                  int similarity_threshold=10, int max_interval=10,
		  int bitmap=0, Set *bitmaps=NULL) {
      sen_select_optarg optarg;
      int i, m, vector[sizeof(int) * 8];
      optarg.mode = mode;
      if (bitmap) {
        for (i = 0, m = 1; i < sizeof(int) * 8; i++, m <<= 1) {
	  vector[i] = m & bitmap ? 1 : 0;
        }
	optarg.weight_vector = vector;
	optarg.vector_size = sizeof(int) * 8;
      } else {
	optarg.weight_vector = NULL;
	optarg.vector_size = 0;
      }
      if (bitmaps) {
        optarg.func = sample_select_func;
        optarg.func_arg = (void *)bitmaps;
      } else {
        optarg.func = NULL;
	optarg.func_arg = NULL;
      }
      return sen_index_select(self, string, r, op, &optarg);
    }
    sen_rc check() {
      return sen_index_check(self);
    }
    static sen_rc remove(const char *path) {
      return sen_index_remove(path);
    }
  }
} Index;

typedef struct {
  %extend {
    sen_key next() {
      sen_key key;
      key.key = sen_records_next(self);
      sen_sym_info(self->keys, &key.size, NULL, NULL, NULL);
      return key;
    }
    sen_rc rewind() {
      return sen_records_rewind(self);
    }
    sen_key curr_key() {
      sen_key key;
      key.key = sen_records_curr_key(self);
      sen_sym_info(self->keys, &key.size, NULL, NULL, NULL);
      return key;
    }
    Record *curr_rec() {
      Record *r = malloc(sizeof(Record));
      if (!r) { SWIG_exception_(SWIG_MemoryError, "memory exhausted"); }
      r->records = self;
      r->rh = self->curr_rec;
      return r;
    }
    Record *at(sen_key key, unsigned section=0, unsigned pos=0) {
      const sen_recordh *rh;
      int key_size;
      sen_sym_info(self->keys, &key_size, NULL, NULL, NULL);
      VALIDATE_KEY_SIZE(key,key_size);
      if ((rh = sen_records_at(self, key.key, section, pos, NULL))) {
        Record *r = malloc(sizeof(Record));
	r->records = self;
	r->rh = rh;
	return r;
      }
      return NULL;
    }
    int curr_score() {
      return sen_records_curr_score(self);
    }
    int nhits() {
      return sen_records_nhits(self);
    }
    int find(sen_key key) {
      int key_size;
      sen_sym_info(self->keys, &key_size, NULL, NULL, NULL);
      VALIDATE_KEY_SIZE(key,key_size);
      return sen_records_find(self, key.key);
    }
    Records(sen_rec_unit record_unit=sen_rec_document,
            sen_rec_unit subrec_unit=sen_rec_none,
	    unsigned max_n_subrecs=0) {
      return sen_records_open(record_unit, subrec_unit, max_n_subrecs);
    }
    ~Records() {
      sen_records_close(self);
    }
    sen_rc sort(int limit, sen_sort_mode mode=sen_sort_descending, bool func=0) {
      sen_sort_optarg optarg;
      optarg.mode = mode;
      optarg.compar = func ? sample_sort_func : NULL;
      optarg.compar_arg = NULL;
      return sen_records_sort(self, limit, &optarg);
    }
    sen_rc group(int limit, sen_sort_mode mode=sen_sort_descending, bool func=0) {
      sen_group_optarg optarg;
      optarg.mode = mode;
      optarg.func = func ? sample_group_func : NULL;
      optarg.func_arg = NULL;
      optarg.key_size = sizeof(int);
      return sen_records_group(self, limit, &optarg);
    }
    Records *concat(Records *r) {
      return sen_records_union(self, r);
    }
    Records *subtract(Records *r) {
      return sen_records_subtract(self, r);
    }
    Records *intersect(Records *r) {
      return sen_records_intersect(self, r);
    }
    int difference(Records *r) {
      return sen_records_difference(self, r);
    }
  }
} Records;

typedef struct {
  %extend {
    Values() {
      return sen_values_open();
    }
    ~Values() {
      sen_values_close(self);
    }
    sen_rc add(const char *str, unsigned weight=0) {
      return sen_values_add(self, str, weight);
    }
  }
} Values;

%apply int *OUTPUT { int *score, int *section, int *pos, int *n_subrecs };

typedef struct {
  %extend {
    sen_key info(int *score, int *section, int *pos, int *n_subrecs) {
      sen_key key;
      sen_recinfo ri;
      sen_rc rc = sen_record_info(self->records, self->rh, &ri);
      RC_HANDLER(rc);
      if (ri.key) {
        key.key = ri.key;
        sen_sym_info(self->records->keys, &key.size, NULL, NULL, NULL);
      } else {
        key.key = ri.userkey;
        key.size = self->records->record_size;
      }
      *score = ri.score;
      *section = ri.section;
      *pos = ri.pos;
      *n_subrecs = ri.n_subrecs;
      return key;
    }
    sen_key subrec_info(int n, int *score, int *section, int *pos, int *n_subrecs) {
      sen_key key;
      sen_recinfo ri;
      int nr = sen_record_subrec_info(self->records, self->rh, &ri, n, 1);
      if (nr != 1) { SWIG_exception_(SWIG_RuntimeError, "other error"); }
      if (ri.key) {
        key.key = ri.key;
        sen_sym_info(self->records->keys, &key.size, NULL, NULL, NULL);
      } else {
        key.key = ri.userkey;
        key.size = self->records->record_size;
      }
      *score = ri.score;
      *section = ri.section;
      *pos = ri.pos;
      *n_subrecs = ri.n_subrecs;
      return key;
    }
  }
} Record;

%apply int *OUTPUT { unsigned *key_size, unsigned *value_size, unsigned *n_entries };

typedef struct {
  %extend {
    Set(unsigned key_size=0, unsigned value_size=0, unsigned initial_size=0) {
      return sen_set_open(key_size, value_size, initial_size);
    }
    ~Set() {
      sen_set_close(self);
    }
    void info(unsigned *key_size, unsigned *value_size, unsigned *n_entries) {
      sen_set_info(self, key_size, value_size, n_entries);
    }
    Set *concat(Set *s) {
      return sen_set_union(self, s);
    }
    Set *subtract(Set *s) {
      return sen_set_subtract(self, s);
    }
    Set *intersect(Set *s) {
      return sen_set_intersect(self, s);
    }
    int difference(Set *s) {
      return sen_set_difference(self, s);
    }
    bool put(sen_key key, sen_key value) {
      void *vp;
      unsigned key_size, value_size;
      sen_set_info(self, &key_size, &value_size, NULL);
      VALIDATE_KEY_SIZE(key,key_size);
      if (sen_set_get(self, key.key, &vp)) {
        int l = value.size ? value.size : strlen(value.key);
	if (l > value_size) {
          memcpy(vp, value.key, value_size);
        } else {
	  memcpy(vp, value.key, l);
	  memset(vp + l, 0, value_size - l);
        }
	return 1;
      } else {
        return 0;
      }
    }
    sen_key at(sen_key key) {
      sen_key value;
      unsigned key_size;
      sen_set_info(self, &key_size, &value.size, NULL);
      VALIDATE_KEY_SIZE(key,key_size);
      if (sen_set_at(self, key.key, (void **)&value.key)) {
        if (!value.size) { value.key = ""; }
      } else {
        value.key = NULL;
      }
      return value;
    }
    sen_rc del(sen_key key) {
      sen_rc rc;
      sen_set_eh *eh;
      unsigned key_size;
      sen_set_info(self, &key_size, NULL, NULL);
      VALIDATE_KEY_SIZE(key,key_size);
      if ((eh = sen_set_at(self, key.key, NULL))) {
        rc = sen_set_del(self, eh);
      } else {
        rc = sen_invalid_argument;
      }
      return rc;
    }
  }
} Set;

%apply sen_key *OUTPUT { sen_key *value };

typedef struct {
  %extend {
    SetCursor(Set *s) {
      SetCursor *sc = malloc(sizeof(SetCursor));
      if (!sc) { SWIG_exception_(SWIG_MemoryError, "memory exhausted"); }
      sc->set = s;
      sc->cursor = sen_set_cursor_open(s);
      return sc;
    }
    ~SetCursor() {
      sen_set_cursor_close(self->cursor);
      free(self);
    }
    sen_key next(sen_key *value) {
      sen_key key;
      sen_set_info(self->set, &key.size, &value->size, NULL);
      if (sen_set_cursor_next(self->cursor, (void **)&key.key, (void **)&value->key)) {
        if (!value->size) { value->key = ""; }
      } else {
        key.key = NULL;
        value->key = NULL;
      }
      return key;
    }
  }
} SetCursor;

typedef struct {
  %extend {
    Sym(const char *path) {
      Sym *res = sen_sym_open(path);
      if (!res) { SWIG_exception_(SWIG_IOError, "file open error"); }
      return res;
    }
    static Sym *open(const char *path) {
      Sym *res = sen_sym_open(path);
      if (!res) { SWIG_exception_(SWIG_IOError, "file open error"); }
      return res;
    }
    static Sym *create(const char *path, unsigned key_size=0,
                       unsigned flags=0, sen_encoding encoding=0) {
      Sym *res = sen_sym_create(path, key_size, flags, encoding);
      if (!res) { SWIG_exception_(SWIG_IOError, "file create error"); }
      return res;
    }
    int key_size() {
      int res;
      if (sen_sym_info(self, &res, NULL, NULL, NULL)) {
        SWIG_exception_(SWIG_UnknownError, "other error");
      }
      return res;
    }
    unsigned flags() {
      unsigned res;
      if (sen_sym_info(self, NULL, &res, NULL, NULL)) {
        SWIG_exception_(SWIG_UnknownError, "other error");
      }
      return res;
    }
    sen_encoding encoding() {
      sen_encoding res;
      if (sen_sym_info(self, NULL, NULL, &res, NULL)) {
        SWIG_exception_(SWIG_UnknownError, "other error");
      }
      return res;
    }
    unsigned nrecords() {
      unsigned res;
      if (sen_sym_info(self, NULL, NULL, NULL, &res)) {
        SWIG_exception_(SWIG_UnknownError, "other error");
      }
      return res;
    }
    ~Sym() {
      sen_sym_close(self);
    }
    sen_id get(const char *key) {
      return sen_sym_get(self, key);
    }
    sen_id at(const char *key) {
      return sen_sym_at(self, key);
    }
    sen_rc del(const char *key) {
      return sen_sym_del(self, key);
    }
    unsigned size() {
      return sen_sym_size(self);
    }
    const char *key(sen_id id) {
      return sen_sym_key(self, id);
    }
    Set *prefix_search(const char *key) {
      return sen_sym_prefix_search(self, key);
    }
    Set *suffix_search(const char *key) {
      return sen_sym_suffix_search(self, key);
    }
    sen_id common_prefix_search(const char *key) {
      return sen_sym_common_prefix_search(self, key);
    }
    static sen_rc remove(const char *path) {
      return sen_sym_remove(path);
    }
  }
} Sym;
