/* Copyright(C) 2004 Brazil

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
  
  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/resource.h>
#include "sym.h"
#include "inv.h"
#include "str.h"
#include "set.h"
#include "lex.h"
#include "senna_in.h"

/* private classes */

/* b-heap */

typedef struct {
  int n_entries;
  int n_bins;
  sen_inv_cursor **bins;
} cursor_heap;

static inline cursor_heap *
cursor_heap_open(int max)
{
  cursor_heap *h = SEN_MALLOC(sizeof(cursor_heap));
  if (!h) { return NULL; }
  h->bins = SEN_MALLOC(sizeof(sen_inv_cursor *) * max);
  if (!h->bins) { 
    SEN_FREE(h);
    return NULL;
  }
  h->n_entries = 0;
  h->n_bins = max;
  return h;
}

static inline sen_rc
cursor_heap_push(cursor_heap *h, sen_inv *inv, sen_id tid, uint32_t offset2)
{
  int n, n2;
  sen_inv_cursor *c, *c2;
  if (h->n_entries >= h->n_bins) {
    int max = h->n_bins * 2;
    sen_inv_cursor **bins = SEN_REALLOC(h->bins, sizeof(sen_inv_cursor *) * max);
    sen_log("expanded cursor_heap to %d,%p", max, bins);
    if (!bins) { return sen_memory_exhausted; }
    h->n_bins = max;
    h->bins = bins;
  }
  if (!(c = sen_inv_cursor_open(inv, tid))) {
    sen_log("cursor open failed");
    return sen_other_error;
  }
  if (sen_inv_cursor_next(c)) {
    sen_inv_cursor_close(c);
    return sen_other_error;
  }
  if (sen_inv_cursor_next_pos(c)) {
    sen_log("invalid inv_cursor b");
    sen_inv_cursor_close(c);
    return sen_other_error;
  }
  // c->offset = offset2;
  n = h->n_entries++;
  while (n) {
    n2 = (n - 1) >> 1;
    c2 = h->bins[n2];
    if (SEN_INV_CURSOR_CMP(c, c2)) { break; }
    h->bins[n] = c2;
    n = n2;
  }
  h->bins[n] = c;
  return sen_success;
}

static inline sen_inv_cursor *
cursor_heap_min(cursor_heap *h)
{
  return h->n_entries ? h->bins[0] : NULL;
}

static inline void
cursor_heap_recalc_min(cursor_heap *h)
{
  int n = 0, n1, n2, m;
  if ((m = h->n_entries) > 1) {
    sen_inv_cursor *c = h->bins[0], *c1, *c2;
    for (;;) {
      n1 = n * 2 + 1;
      n2 = n1 + 1;
      c1 = n1 < m ? h->bins[n1] : NULL;
      c2 = n2 < m ? h->bins[n2] : NULL;
      if (c1 && SEN_INV_CURSOR_CMP(c, c1)) {
	if (c2 && SEN_INV_CURSOR_CMP(c, c2) && SEN_INV_CURSOR_CMP(c1, c2)) {
	  h->bins[n] = c2;
	  n = n2;
	} else {
	  h->bins[n] = c1;
	  n = n1;
	}
      } else {
	if (c2 && SEN_INV_CURSOR_CMP(c, c2)) {
	  h->bins[n] = c2;
	  n = n2;
	} else {
	  h->bins[n] = c;
	  break;
	}
      }
    }
  }
}

static inline void
cursor_heap_pop(cursor_heap *h)
{
  if (h->n_entries) {
    sen_inv_cursor *c = h->bins[0];
    if (sen_inv_cursor_next(c)) {
      sen_inv_cursor_close(c);
      h->bins[0] = h->bins[--h->n_entries];
    } else if (sen_inv_cursor_next_pos(c)) {
      sen_log("invalid inv_cursor c");
      sen_inv_cursor_close(c);
      h->bins[0] = h->bins[--h->n_entries];
    }
    cursor_heap_recalc_min(h);
  }
}

static inline void
cursor_heap_pop_pos(cursor_heap *h)
{
  if (h->n_entries) {
    sen_inv_cursor *c = h->bins[0];
    if (sen_inv_cursor_next_pos(c)) {
      if (sen_inv_cursor_next(c)) {
	sen_inv_cursor_close(c);
	h->bins[0] = h->bins[--h->n_entries];
      } else if (sen_inv_cursor_next_pos(c)) {
	sen_log("invalid inv_cursor d");
	sen_inv_cursor_close(c);
	h->bins[0] = h->bins[--h->n_entries];
      }
    }
    cursor_heap_recalc_min(h);
  }
}

static inline void
cursor_heap_close(cursor_heap *h)
{
  int i = h->n_entries;
  while (i--) { sen_inv_cursor_close(h->bins[i]); }
  SEN_FREE(h->bins);
  SEN_FREE(h);
}

/* token_info */

typedef struct {
  cursor_heap *cursors;
  uint32_t rid;
  uint32_t sid;
  int offset;
  int pos;
  int size;
  int score;
  int ntoken;
} token_info;

typedef enum {
  expand_none = 0,
  expand_prefix,
  expand_suffix,
  expand_both
} expand_mode;

inline static void
token_info_expand_both(sen_index *i, const char *key, token_info *ti)
{
  int s = 0;
  sen_set *h, *g;
  uint32_t *offset2;
  sen_set_cursor *c;
  sen_id *tp, *tq;
  if (!(h = sen_sym_prefix_search(i->lexicon, key))) { return; }
  sen_log("key=%s h->n=%d", key, h->n_entries);
  ti->cursors = cursor_heap_open(h->n_entries + 256);
  c = sen_set_cursor_open(h);
  while (sen_set_cursor_next(c, (void **) &tp, NULL)) {
    const char *key2 = sen_sym_key(i->lexicon, *tp);
    // sen_log("key2=%s", key2);
    if (sen_str_len(key2, i->lexicon->encoding, NULL) == 1) { // todo: 
      if ((s = sen_inv_estimate_size(i->inv, *tp))) {
	cursor_heap_push(ti->cursors, i->inv, *tp, 0);
	ti->ntoken++;
	ti->size += s;
      }
    } else {
      g = sen_sym_suffix_search(i->lexicon, key2);
      // sen_log("g=%d", g ? g->n_entries : 0);
      if (g) {
	sen_set_cursor *d = sen_set_cursor_open(g);
	while (sen_set_cursor_next(d, (void **) &tq, (void **) &offset2)) {
	  if ((s = sen_inv_estimate_size(i->inv, *tq))) {
	    // sen_log("est=%d key=%s", s, sen_sym_key(i->lexicon, *tq));
	    cursor_heap_push(ti->cursors, i->inv, *tq, /* *offset2 */ 0);
	    ti->ntoken++;
	    ti->size += s;
	  }
	}
	sen_set_cursor_close(d);
	sen_set_close(g);
      }
    }
    // if (ti->size > 1000000) { sen_log("huge! %d", ti->size); break; } // todo!!!
  }
  sen_set_cursor_close(c);
  sen_set_close(h);
  sen_log("key=%s done..", key);
}

inline static sen_rc
token_info_close(token_info *ti)
{
  cursor_heap_close(ti->cursors);
  SEN_FREE(ti);
  return sen_success;
}

inline static token_info *
token_info_open(sen_index *i, const char *key, uint32_t offset, expand_mode mode)
{
  int s = 0;
  sen_set *h;
  token_info *ti;
  sen_id tid;
  sen_set_cursor *c;
  sen_id *tp;
  if (!(ti = SEN_MALLOC(sizeof(token_info)))) { return NULL; }
  ti->cursors = NULL;
  ti->size = 0;
  ti->ntoken = 0;
  ti->offset = offset;
  switch (mode) {
  case expand_both :
    token_info_expand_both(i, key, ti);
    break;
  case expand_none :
    if (!(tid = sen_sym_at(i->lexicon, key))) { break; }
    if (!(s = sen_inv_estimate_size(i->inv, tid))) { break; }
    ti->cursors = cursor_heap_open(1);
    cursor_heap_push(ti->cursors, i->inv, tid, 0);
    ti->ntoken++;
    ti->size = s;
    break;
  case expand_prefix :
    if (!(h = sen_sym_prefix_search(i->lexicon, key))) { break; }
    sen_log("key=%s h->n=%d", key, h->n_entries);
    ti->cursors = cursor_heap_open(h->n_entries);
    c = sen_set_cursor_open(h);
    while (sen_set_cursor_next(c, (void **) &tp, NULL)) {
      if ((s = sen_inv_estimate_size(i->inv, *tp))) {
	// sen_log("%8d %s", s, sen_sym_key(i->lexicon, *tp));
	cursor_heap_push(ti->cursors, i->inv, *tp, 0);
	ti->ntoken++;
	ti->size += s;
      }
      //if (ti->size > 1000000) { sen_log("huge! %d", ti->size); break; } // todo!!!
    }
    sen_set_cursor_close(c);
    sen_set_close(h);
    sen_log("key=%s done.", key);
    break;
  case expand_suffix :
    {
      uint32_t *offset2;
      if (!(h = sen_sym_suffix_search(i->lexicon, key))) { break; }
      sen_log("key=%s h->n=%d", key, h->n_entries);
      ti->cursors = cursor_heap_open(h->n_entries);
      c = sen_set_cursor_open(h);
      while (sen_set_cursor_next(c, (void **) &tp, (void **) &offset2)) {
	if ((s = sen_inv_estimate_size(i->inv, *tp))) {
	  cursor_heap_push(ti->cursors, i->inv, *tp, /* *offset2 */ 0);
	  ti->ntoken++;
	  ti->size += s;
	}
      }
      sen_set_cursor_close(c);
      sen_set_close(h);
      break;
    }
  }
  {
    sen_inv_cursor *ic;
    if (ti->cursors && (ic = cursor_heap_min(ti->cursors))) {
      sen_inv_posting *p = ic->post;
      ti->rid = p->rid;
      ti->sid = p->sid;
      ti->pos = p->pos;
      ti->score = p->score;
    } else {
      token_info_close(ti);
      ti = NULL;
    }
  }
  return ti;
}

static inline sen_rc
token_info_skip(token_info *ti, uint32_t rid, uint32_t sid)
{
  sen_inv_cursor *c;
  sen_inv_posting *p;
  for (;;) {
    if (!(c = cursor_heap_min(ti->cursors))) { return sen_other_error; }
    p = c->post;
    if (p->rid > rid || (p->rid == rid && p->sid >= sid)) { break; }
    cursor_heap_pop(ti->cursors);
  }
  // sen_log("r=%d s=%d pr=%d ps=%d", rid, sid, p->rid, p->sid);
  ti->rid = p->rid;
  ti->sid = p->sid;
  ti->pos = p->pos - ti->offset;
  ti->score = p->score;
  return sen_success;
}

static inline sen_rc
token_info_skip_pos(token_info *ti, uint32_t rid, uint32_t sid, uint32_t pos)
{
  sen_inv_cursor *c;
  sen_inv_posting *p;
  // sen_log("r=%d s=%d p=%d o=%d", rid, sid, pos, ti->offset);
  pos += ti->offset;
  for (;;) {
    if (!(c = cursor_heap_min(ti->cursors))) { return sen_other_error; }
    p = c->post;
    if (p->rid != rid || p->sid != sid || p->pos >= pos) { break; }
    cursor_heap_pop_pos(ti->cursors);
  }
  ti->rid = p->rid;
  ti->sid = p->sid;
  ti->pos = p->pos - ti->offset;
  ti->score = p->score;
  return sen_success;
}

inline static int
token_compare(const void *a, const void *b)
{
  const token_info *t1 = *((token_info **)a), *t2 = *((token_info **)b);
  return t1->size - t2->size;
}

inline static sen_rc
token_info_build(sen_index *i, const char *string, token_info **tis, uint32_t *n,
		 sen_sel_mode mode)
{
  token_info *ti;
  sen_rc rc = sen_other_error;
  if (mode == sen_sel_unsplit) {
    if ((ti = token_info_open(i, string, 0, expand_both))) {
      tis[(*n)++] = ti;
      rc = sen_success;
    }
  } else {
    sen_id tid;
    int partialp = mode == sen_sel_partial;
    sen_lex *lex = sen_lex_open(i->lexicon, string, 0);
    if (!lex) { return sen_memory_exhausted; }
    tid = sen_lex_next(lex);
    switch (lex->status) {
    case sen_lex_doing :
      ti = token_info_open(i, sen_sym_key(i->lexicon, tid), lex->pos,
			   partialp ? expand_suffix : expand_none);
      break;
    case sen_lex_done :
      ti = token_info_open(i, sen_sym_key(i->lexicon, tid), lex->pos,
			   partialp ? expand_both : expand_none);
      break;
    case sen_lex_not_found :
      ti = token_info_open(i, lex->str, 0,
			   partialp ? expand_both : expand_none);
      break;
    default :
      goto exit;
    }
    if (!ti) { goto exit ; }
    tis[(*n)++] = ti;
    sen_log("%d:%s(%d)", lex->pos, (tid == SEN_SYM_NIL) ? lex->str : sen_sym_key(i->lexicon, tid), tid);
    while (lex->status == sen_lex_doing) {
      tid = sen_lex_next(lex);
      switch (lex->status) {
      case sen_lex_doing :
	ti = token_info_open(i, sen_sym_key(i->lexicon, tid), lex->pos, expand_none);
	break;
      case sen_lex_done :
	ti = token_info_open(i, sen_sym_key(i->lexicon, tid), lex->pos,
			     partialp ? expand_prefix : expand_none);
	break;
      default :
	ti = token_info_open(i, lex->next, lex->pos,
			     partialp ? expand_prefix : expand_none);
	break;
      }
      if (!ti) { goto exit; }
      tis[(*n)++] = ti;
      sen_log("%d:%s(%d)", lex->pos, (tid == SEN_SYM_NIL) ? lex->next : sen_sym_key(i->lexicon, tid), tid);
    }
    rc = sen_success;
exit :
    sen_lex_close(lex);
  }
  return rc;
}

/* senna.conf */

#define CONF_PATH SENNA_HOME "/senna.conf"

static int sen_partial_match_threshold = 0;
static uint32_t sen_initial_n_segments = 0;
static sen_encoding sen_default_encoding = sen_enc_euc_jp;
static char *initial_n_segments = "INITIAL_N_SEGMENTS";
static char *default_encoding = "DEFAULT_ENCODING";
static char *partial_match_threshold = "PARTIAL_MATCH_THRESHOLD";

#define BUFSIZE 32768

static void 
read_conf(void)
{
  FILE *fp;
  char *buffer, *p;
  if (!(fp = fopen(CONF_PATH, "r"))) { return; }
  if (!(buffer = SEN_MALLOC(BUFSIZE))) { return ; }
  while (!feof(fp)) {
    if (!fgets(buffer, BUFSIZE, fp)) { break; }
    if (!strncmp(buffer, initial_n_segments, strlen(initial_n_segments))) {
      for (p = buffer; *p && (*p < '0' || *p > '9'); p++) {
	if (p - buffer > BUFSIZE) { break; }
      }
      sen_initial_n_segments = atoi(p);
      sen_log("INITIAL_N_SEGMENTS=%d", sen_initial_n_segments);
    }
    if (!strncmp(buffer, partial_match_threshold, strlen(partial_match_threshold))) {
      for (p = buffer; *p && (*p < '0' || *p > '9') && *p != '-'; p++) {
	if (p - buffer > BUFSIZE) { break; }
      }
      sen_partial_match_threshold = atoi(p);
      sen_log("PARTIAL_MATCH_THRESHOLD=%d", sen_partial_match_threshold);
    }
    if (!strncmp(buffer, default_encoding, strlen(default_encoding))) {
      for (p = buffer + strlen(default_encoding); *p; p++) {
	if (p - buffer > BUFSIZE) { break; }
	if (*p == 'e' || *p == 'E') {
	  sen_default_encoding = sen_enc_euc_jp;
	  break;
	}
	if (*p == 'u' || *p == 'U') {
	  sen_default_encoding = sen_enc_utf8;
	  break;
	}
	if (*p == 's' || *p == 'S') {
	  sen_default_encoding = sen_enc_sjis;
	  break;
	}
	if (*p == 'n' || *p == 'N') {
	  sen_default_encoding = sen_enc_none;
	  break;
	}
      }
    }
  }
  SEN_FREE(buffer);
  fclose(fp);
  {
  struct rlimit rlim;
  if (!getrlimit(RLIMIT_STACK, &rlim)) {
    sen_log("RLIMIT_STACK is %d (%d)", rlim.rlim_cur, getpid());
    if (rlim.rlim_cur != RLIM_INFINITY && rlim.rlim_cur < SEN_STACK_SIZE) {
      rlim.rlim_cur = SEN_STACK_SIZE;
      if (!setrlimit(RLIMIT_STACK, &rlim)) {
	sen_log("expanded RLIMIT_STACK to %d", rlim.rlim_cur);
      }
    }
  }
  }
}

/* init */

sen_rc
sen_init(void)
{
  struct rlimit rlim;
  if (!getrlimit(RLIMIT_STACK, &rlim)) {
    sen_log("RLIMIT_STACK is %d", rlim.rlim_cur);
    if (rlim.rlim_cur != RLIM_INFINITY && rlim.rlim_cur < SEN_STACK_SIZE) {
      rlim.rlim_cur = SEN_STACK_SIZE;
      if (!setrlimit(RLIMIT_STACK, &rlim)) {
	sen_log("expanded RLIMIT_STACK to %d", rlim.rlim_cur);
      }
    }
  }
  return sen_lex_init();
}

sen_rc
sen_fin(void)
{
  return sen_str_fin();
}

/* sen_index */

sen_index *
sen_index_create(const char *path, int key_size,
		 int flags, int initial_n_segments, sen_encoding encoding)
{
  sen_index *i;
  char buffer[PATH_MAX];
  read_conf();
  if (initial_n_segments == 0) { initial_n_segments = sen_initial_n_segments; }
  if (encoding == sen_enc_default) { encoding = sen_default_encoding; }
  if (strlen(path) > PATH_MAX - 4) {
    sen_log("too long index path (%s)", path);
    return NULL;
  }
  if (!(i = SEN_MALLOC(sizeof(sen_index)))) { return NULL; }
  if (encoding == sen_enc_utf8) {
    flags &= ~SEN_INDEX_NORMALIZE;
    /* todo: SEN_INDEX_NORMALIZE is not supported with utf8 */
  }
  sen_log("creating '%s' encoding=%s initial_n_segments=%d",
	  path, sen_enc_string[encoding], initial_n_segments);
  strcpy(buffer, path);
  strcat(buffer, ".SEN");
  if ((i->keys = sen_sym_create(buffer, key_size, 0, sen_enc_none))) {
    strcpy(buffer, path);
    strcat(buffer, ".SEN.l");
    if ((i->lexicon = sen_sym_create(buffer, 0, SEN_SYM_WITH_SIS|flags, encoding))) {
      strcpy(buffer, path);
      strcat(buffer, ".SEN.i");
      if ((i->inv = sen_inv_create(buffer, i->lexicon, 0, initial_n_segments))) {
	// pthread_mutex_init(&i->write_mutex, NULL);
	sen_log("index created (%s) flags=%x", path, i->lexicon->flags);
	return i;
      }
      sen_sym_close(i->lexicon);
    }
    sen_sym_close(i->keys);
  }
  SEN_FREE(i);
  return NULL;
}

sen_index *
sen_index_open(const char *path)
{
  sen_index *i;
  char buffer[PATH_MAX];
  if (strlen(path) > PATH_MAX - 4) {
    sen_log("too long index path (%s)", path);
    return NULL;
  }
  read_conf();
  if (!(i = SEN_MALLOC(sizeof(sen_index)))) { return NULL; }
  strcpy(buffer, path);
  strcat(buffer, ".SEN");
  if ((i->keys = sen_sym_open(buffer))) {
    strcpy(buffer, path);
    strcat(buffer, ".SEN.l");
    if ((i->lexicon = sen_sym_open(buffer))) {
      strcpy(buffer, path);
      strcat(buffer, ".SEN.i");
      if ((i->inv = sen_inv_open(buffer, i->lexicon))) {
	// pthread_mutex_init(&i->write_mutex, NULL);
	sen_log("index opened (%p:%s) flags=%x", i, path, i->lexicon->flags);
	return i;
      }
      sen_sym_close(i->lexicon);
    }
    sen_sym_close(i->keys);
  }
  SEN_FREE(i);
  return NULL;
}

sen_index *
sen_index_create_with_keys(const char *path, sen_sym *keys,
			   int flags, int initial_n_segments, sen_encoding encoding)
{
  sen_index *i;
  char buffer[PATH_MAX];
  read_conf();
  if (initial_n_segments == 0) { initial_n_segments = sen_initial_n_segments; }
  if (encoding == sen_enc_default) { encoding = sen_default_encoding; }
  if (strlen(path) > PATH_MAX - 4) {
    sen_log("too long index path (%s)", path);
    return NULL;
  }
  if (!(i = SEN_MALLOC(sizeof(sen_index)))) { return NULL; }
  if (encoding == sen_enc_utf8) {
    flags &= ~SEN_INDEX_NORMALIZE;
    /* todo: SEN_INDEX_NORMALIZE is not supported with utf8 */
  }
  sen_log("creating '%s' encoding=%s initial_n_segments=%d",
	  path, sen_enc_string[encoding], initial_n_segments);
  i->keys = keys;
  strcpy(buffer, path);
  strcat(buffer, ".SEN.l");
  if ((i->lexicon = sen_sym_create(buffer, 0, SEN_SYM_WITH_SIS|flags, encoding))) {
    strcpy(buffer, path);
    strcat(buffer, ".SEN.i");
    if ((i->inv = sen_inv_create(buffer, i->lexicon, 0, initial_n_segments))) {
      // pthread_mutex_init(&i->write_mutex, NULL);
      sen_log("index created (%s) flags=%x", path, i->lexicon->flags);
      return i;
    }
    sen_sym_close(i->lexicon);
  }
  SEN_FREE(i);
  return NULL;
}

sen_index *
sen_index_open_with_keys(const char *path, sen_sym *keys)
{
  sen_index *i;
  char buffer[PATH_MAX];
  if (strlen(path) > PATH_MAX - 4) {
    sen_log("too long index path (%s)", path);
    return NULL;
  }
  read_conf();
  if (!(i = SEN_MALLOC(sizeof(sen_index)))) { return NULL; }
  i->keys = keys;
  strcpy(buffer, path);
  strcat(buffer, ".SEN.l");
  if ((i->lexicon = sen_sym_open(buffer))) {
    strcpy(buffer, path);
    strcat(buffer, ".SEN.i");
    if ((i->inv = sen_inv_open(buffer, i->lexicon))) {
      // pthread_mutex_init(&i->write_mutex, NULL);
      sen_log("index opened (%s) flags=%x", path, i->lexicon->flags);
      return i;
    }
    sen_sym_close(i->lexicon);
  }
  SEN_FREE(i);
  return NULL;
}

sen_rc
sen_index_close(sen_index *i)
{
  if (!i) { return sen_invalid_argument; }
  // pthread_mutex_destroy(&i->write_mutex);
  sen_sym_close(i->keys);
  sen_sym_close(i->lexicon);
  sen_inv_close(i->inv);
  SEN_FREE(i);
  return sen_success;
}

sen_rc
sen_index_remove(const char *path)
{
  char buffer[PATH_MAX];
  if (!path || strlen(path) > PATH_MAX - 8) { return sen_invalid_argument; }
  snprintf(buffer, PATH_MAX, "%s.SEN", path);
  unlink(buffer);
  snprintf(buffer, PATH_MAX, "%s.SEN.i", path);
  unlink(buffer);
  /* todo: SEN.i.c.0n file may be created */
  snprintf(buffer, PATH_MAX, "%s.SEN.i.c", path);
  unlink(buffer);
  snprintf(buffer, PATH_MAX, "%s.SEN.l", path);
  unlink(buffer);
  return sen_success;
}

sen_rc
sen_index_info(sen_index *i, int *key_size, int *flags,
	       int *initial_n_segments, sen_encoding *encoding)
{
  if (!i) { return sen_invalid_argument; }
  if (key_size) { *key_size = i->keys->key_size; }
  if (flags) { *flags = i->lexicon->flags & ~SEN_SYM_WITH_SIS; }
  if (initial_n_segments) { *initial_n_segments = sen_inv_initial_n_segments(i->inv); }
  if (encoding) { *encoding = i->lexicon->encoding; }
  return sen_success;
}

sen_rc
sen_index_check(sen_index *i)
{
  return sen_inv_check(i->inv);
}

/* update */

sen_rc
sen_index_add(sen_index *i, const void *key, const char *value)
{
  int hint;
  sen_set *h;
  sen_lex *lex;
  sen_set_cursor *c;
  sen_inv_updspec **u;
  sen_id rid, tid, *tp;

  // pthread_mutex_lock(&i->write_mutex);

  sen_log("add > (%x)", key);
  rid = sen_sym_get(i->keys, key);
  h = sen_set_open(sizeof(sen_id), sizeof(sen_inv_updspec *), 0);
  lex = sen_lex_open(i->lexicon, value, 1);
  while (!lex->status) {
    if ((tid = sen_lex_next(lex))) {
      sen_set_get(h, &tid, (void **) &u);
      if (!*u) { *u = sen_inv_updspec_open(rid, 1); }
      sen_inv_updspec_add(*u, lex->pos, 0);
    }
  }
  sen_lex_close(lex);
  c = sen_set_cursor_open(h);
  while (sen_set_cursor_next(c, (void **)&tp, (void **) &u)) {
    hint = sen_str_get_prefix_order(sen_sym_key(i->lexicon, *tp));
    if (hint == -1) { hint = *tp; }
    //      sen_log("inv_update > %d '%s'", *tp, sen_sym_key(i->lexicon, *tp));
    sen_inv_update(i->inv, *tp, *u, hint);
    //      sen_log("inv_update < %d '%s'", *tp, sen_sym_key(i->lexicon, *tp));
    // sen_log("(%s)", sen_sym_key(i->lexicon, *tp));
    sen_inv_updspec_close(*u);
  }
  // pthread_mutex_unlock(&i->write_mutex);
  sen_set_cursor_close(c);
  sen_set_close(h);
  sen_log("add < (%x:%d)", key, rid);
  return sen_success;
}

sen_rc
sen_index_del(sen_index *i, const void *key, const char *value)
{
  sen_set *h;
  sen_lex *lex;
  sen_set_cursor *c;
  sen_inv_updspec **u;
  sen_id rid, tid, *tp;
  //  sen_log("deleting(%s)", value);
  if (!(rid = sen_sym_at(i->keys, key))) { 
    sen_log("del : (%x) (invalid key)", key);
    return sen_invalid_argument;
  }
  sen_log("del > (%x:%d)", key, rid);
  h = sen_set_open(sizeof(sen_id), sizeof(sen_inv_updspec *), 0);
  lex = sen_lex_open(i->lexicon, value, 0);
  while (!lex->status) {
    if ((tid = sen_lex_next(lex))) {
      sen_set_get(h, &tid, (void **) &u);
      if (!*u) { *u = sen_inv_updspec_open(rid, 0); }
    }
  }
  sen_lex_close(lex);
  c = sen_set_cursor_open(h);
  // pthread_mutex_lock(&i->write_mutex);
  while (sen_set_cursor_next(c, (void **)&tp, (void **) &u)) {
    if (*tp) {
      //      sen_log("inv_delete > %d '%s'", *tp, sen_sym_key(i->lexicon, *tp));
      sen_inv_delete(i->inv, *tp, *u);
      //      sen_log("inv_delete < %d '%s'", *tp, sen_sym_key(i->lexicon, *tp));
    }
    sen_inv_updspec_close(*u);
  }
  // pthread_mutex_unlock(&i->write_mutex);
  sen_set_cursor_close(c);
  sen_set_close(h);
  sen_log("del < (%x)", key);
  return sen_success;
}

sen_rc 
sen_index_upd(sen_index *i, const void *key,
	      const char *oldvalue, const char *newvalue)
{
  // if (!strcmp(oldvalue, newvalue)) { return sen_success; }
  sen_rc rc = sen_success;
  // sen_log("index_upd %p %x > '%s'->'%s'", i, key, oldvalue, newvalue);
  if (oldvalue && *oldvalue) { rc = sen_index_del(i, key, oldvalue); }
  if (rc != sen_success) { return rc; }
  if (newvalue && *newvalue) { rc = sen_index_add(i, key, newvalue); }
  //  sen_log("index_upd %x < %d", key, rc);
  return rc;
}

#define INITIAL_VALUE_SIZE 1024

sen_values *
sen_values_open(void)
{
  sen_values *v = SEN_MALLOC(sizeof(sen_values));
  if (v) {
    v->n_values = 0;
    v->values = SEN_MALLOC(sizeof(sen_value) * INITIAL_VALUE_SIZE);
    if (!v->values) {
      SEN_FREE(v);
      v = NULL;
    }
  }
  return v;
}

sen_rc
sen_values_close(sen_values *v)
{
  if (!v) { return sen_invalid_argument; }
  if (v->values) { SEN_FREE(v->values); }
  SEN_FREE(v);
  return sen_success;
}

sen_rc 
sen_values_add(sen_values *v, const char *str, unsigned int weight)
{
  sen_value *vp;
  if (!(v->n_values % INITIAL_VALUE_SIZE)) {
    vp = SEN_REALLOC(v->values, sizeof(sen_value) * (v->n_values + INITIAL_VALUE_SIZE));
    sen_log("expanded values to %d,%p", v->n_values + INITIAL_VALUE_SIZE, vp);
    if (!vp) { return sen_memory_exhausted; }
    v->values = vp;
  }
  vp = &v->values[v->n_values];
  vp->str = str;
  vp->weight = weight;
  v->n_values++;
  return sen_success;
}

sen_rc 
sen_index_update(sen_index *i, const void *key, unsigned int section,
		 sen_values *oldvalues, sen_values *newvalues)
{

  int j, hint;
  sen_value *v;
  sen_set *old, *new;
  sen_lex *lex;
  sen_set_eh *eh;
  sen_set_cursor *c;
  sen_inv_updspec **u, **un;
  sen_id rid, tid, *tp;
  // pthread_mutex_lock(&i->write_mutex);
  if (newvalues) {
    rid = sen_sym_get(i->keys, key);
  } else {
    if (!(rid = sen_sym_at(i->keys, key))) { return sen_invalid_argument; }
  }
  old = sen_set_open(sizeof(sen_id), sizeof(sen_inv_updspec *), 0);
  for (j = oldvalues->n_values, v = oldvalues->values; j; j--, v++) {
    lex = sen_lex_open(i->lexicon, v->str, 0);
    while (!lex->status) {
      if ((tid = sen_lex_next(lex))) {
	sen_set_get(old, &tid, (void **) &u);
	if (!*u) { *u = sen_inv_updspec_open(rid, section); }
	sen_inv_updspec_add(*u, lex->pos, v->weight);
      }
    }
    sen_lex_close(lex);
  }

  new = sen_set_open(sizeof(sen_id), sizeof(sen_inv_updspec *), 0);
  for (j = newvalues->n_values, v = newvalues->values; j; j--, v++) {
    lex = sen_lex_open(i->lexicon, v->str, 1);
    while (!lex->status) {
      if ((tid = sen_lex_next(lex))) {
	sen_set_get(new, &tid, (void **) &u);
	if (!*u) { *u = sen_inv_updspec_open(rid, section); }
	sen_inv_updspec_add(*u, lex->pos, v->weight);
      }
    }
    sen_lex_close(lex);
  }

  c = sen_set_cursor_open(old);
  while (sen_set_cursor_next(c, (void **)&tp, (void **) &u)) {
    if ((eh = sen_set_at(new, tp, (void **) &un))) {
      if (!sen_inv_updspec_cmp(*u, *un)) {
	sen_inv_updspec_close(*un);
	sen_set_del(new, eh);
      }
    } else {
      sen_inv_delete(i->inv, *tp, *u);
    }
    sen_inv_updspec_close(*u);
  }
  sen_set_cursor_close(c);
  sen_set_close(old);

  c = sen_set_cursor_open(new);
  while (sen_set_cursor_next(c, (void **)&tp, (void **) &u)) {
    hint = sen_str_get_prefix_order(sen_sym_key(i->lexicon, *tp));
    if (hint == -1) { hint = *tp; }
    sen_inv_update(i->inv, *tp, *u, hint);
    sen_inv_updspec_close(*u);
  }
  // pthread_mutex_unlock(&i->write_mutex);
  sen_set_cursor_close(c);
  sen_set_close(new);
  return sen_success;
}

/* sen_records */

#define SCORE_SIZE (sizeof(int))

typedef struct {
  int score;
  int n_subrecs;
  int subrecs[1];
} recinfo;

inline static int
rec_unit_size(sen_rec_unit unit)
{
  switch (unit) {
  case sen_rec_document :
    return sizeof(sen_id);
  case sen_rec_section :
    return sizeof(sen_id) + sizeof(int);
  case sen_rec_position :
    return sizeof(sen_id) + sizeof(int) + sizeof(int);
  default :
    return -1;
  }
}

sen_records *
sen_records_open(sen_rec_unit record_unit,
		 sen_rec_unit subrec_unit, unsigned int max_n_subrecs)
{
  sen_records *r;
  int record_size = rec_unit_size(record_unit);
  int subrec_size = rec_unit_size(subrec_unit);
  if (record_size < 0) { return NULL; }
  if (max_n_subrecs && subrec_size <= record_size) { return NULL; }
  if (!(r = SEN_MALLOC(sizeof(sen_records)))) { return NULL; }
  r->record_unit = record_unit;
  r->subrec_unit = subrec_unit;
  r->record_size = record_size;
  r->subrec_size = subrec_size - record_size;
  r->max_n_subrecs = max_n_subrecs;
  r->keys = NULL;
  r->cursor = NULL;
  r->sorted = NULL;
  r->curr_rec = NULL;
  if (!(r->records = sen_set_open(r->record_size,
				  SCORE_SIZE + sizeof(int) +
				  max_n_subrecs * (SCORE_SIZE + r->subrec_size), 0))) {
    SEN_FREE(r);
    return NULL;
  }
  return r;
}

inline static void
sen_records_cursor_clear(sen_records *r)
{
  if (r->sorted) { 
    SEN_FREE(r->sorted);
    r->sorted = NULL;
  }
  if (r->cursor) {
    sen_set_cursor_close(r->cursor);
    r->cursor = NULL;
  }
  r->curr_rec = NULL;
}

sen_rc
sen_records_close(sen_records *r)
{
  if (!r) { return sen_invalid_argument; }
  sen_records_cursor_clear(r);
  sen_set_close(r->records);
  SEN_FREE(r);
  sen_log("sen_records_close < (%x)", r);
  return sen_success;
}

sen_rc 
sen_records_sort(sen_records *r, int limit, sen_sort_optarg *optarg)
{
  sen_set_sort_optarg arg;
  if (!r || !r->records) { return sen_invalid_argument; }
  sen_records_cursor_clear(r);
  if (limit > r->records->n_entries) { limit = r->records->n_entries; }
  if (optarg) {
    arg.mode = optarg->mode;
    arg.compar = (int (*)(sen_set *, sen_set_eh *, sen_set_eh *, void *))optarg->compar;
    arg.compar_arg = optarg->compar_arg;
    arg.compar_arg0 = (sen_set *) r;
    r->sorted = sen_set_sort(r->records, limit, &arg);
  } else {
    arg.mode = sen_sort_descending;
    arg.compar = NULL;
    arg.compar_arg = (void *)r->records->key_size;
    arg.compar_arg0 = NULL;
    r->sorted = sen_set_sort(r->records, limit, &arg);
  }
  r->limit = limit;
  return r->sorted ? sen_success : sen_other_error;
}

sen_rc
sen_records_rewind(sen_records *r)
{
  if (!r) { return sen_invalid_argument; }
  sen_records_cursor_clear(r);
  r->cursor = sen_set_cursor_open(r->records);
  return r->cursor ? sen_success : sen_other_error;
}

const void *
sen_records_next(sen_records *r)
{
  if (!r) { return NULL; }
  if (r->sorted) {
    if (r->curr_rec) {
      if ((++r->curr_rec - r->sorted) >= r->limit) {
	r->curr_rec = NULL;
      }
    } else {
      r->curr_rec = r->sorted;
    }
  } else {
    if (!r->cursor) { sen_records_rewind(r); }
    r->curr_rec = sen_set_cursor_next(r->cursor, NULL, NULL);
  }
  if (r->curr_rec) {
    sen_id *rid;
    if (!sen_set_element_info(r->records, r->curr_rec, (void **)&rid, NULL)) {
      // sen_log("sen_record_next=%d,%d", *rid, *((int *)sen_sym_key(r->keys, *rid)));
      return sen_sym_key(r->keys, *rid);
    }
  }
  return NULL;
}

int
sen_records_find(sen_records *r, const void *key)
{
  sen_id rid;
  recinfo *ri;
  if (!r || !r->keys || r->record_unit != sen_rec_document) { return 0; }
  if (!(rid = sen_sym_at(r->keys, key))) { return 0; }
  if (!sen_set_at(r->records, &rid, (void **) &ri)) { return 0; }
  return ri->score;
}

int
sen_records_nhits(sen_records *r)
{
  if (!r || !r->records) { return 0; }
  return r->records->n_entries;
}

int
sen_records_curr_score(sen_records *r)
{
  recinfo *ri;
  if (!r || !r->curr_rec) { return 0; }
  if (sen_set_element_info(r->records, r->curr_rec, NULL, (void **)&ri)) { return 0; }
  return ri->score;
}

const void *
sen_records_curr_key(sen_records *r)
{
  sen_id *rid;
  if (!r || !r->curr_rec) { return NULL; }
  if (sen_set_element_info(r->records, r->curr_rec, (void **)&rid, NULL)) {
    return NULL;
  }
  return sen_sym_key(r->keys, *rid);
}

sen_records *
sen_records_union(sen_records *a, sen_records *b)
{
  if (a->keys != b->keys) { return NULL; }
  if (!sen_set_union(a->records, b->records)) { return NULL; }
  b->records = NULL;
  sen_records_close(b);
  sen_records_cursor_clear(a);
  return a;
}

sen_records *
sen_records_subtract(sen_records *a, sen_records *b)
{
  if (a->keys != b->keys) { return NULL; }
  if (!sen_set_subtract(a->records, b->records)) { return NULL; }
  b->records = NULL;
  sen_records_close(b);
  sen_records_cursor_clear(a);
  return a;
}

sen_records *
sen_records_intersect(sen_records *a, sen_records *b)
{
  if (a->keys != b->keys) { return NULL; }
  if (!sen_set_intersect(a->records, b->records)) { return NULL; }
  b->records = NULL;
  sen_records_close(b);
  sen_records_cursor_clear(a);
  return a;
}

int 
sen_records_difference(sen_records *a, sen_records *b)
{
  int count;
  if (a->keys != b->keys) { return -1; }
  if ((count = sen_set_difference(a->records, b->records)) >= 0) {
    sen_records_cursor_clear(a);
    sen_records_cursor_clear(b);
  }
  return count;
}

#define SUBRECS_CMP(a,b,dir) (((a) - (b))*(dir) > 0)
#define SUBRECS_NTH(subrecs,size,n) ((int *)(subrecs + n * (size + SCORE_SIZE)))
#define SUBRECS_COPY(subrecs,size,n,src) \
  (memcpy(subrecs + n * (size + SCORE_SIZE), src, size + SCORE_SIZE))

inline static void
subrecs_push(void *subrecs, int size, int n_subrecs, int score, void *body, int dir)
{
  void *v;
  int *c2;
  int n = n_subrecs - 1, n2;
  while (n) {
    n2 = (n - 1) >> 1;
    c2 = SUBRECS_NTH(subrecs,size,n2);
    if (SUBRECS_CMP(score, *c2, dir)) { break; }
    SUBRECS_COPY(subrecs,size,n,c2);
    n = n2;
  }
  v = subrecs + n * (size + SCORE_SIZE);
  *((int *)v) = score;
  memcpy(v + SCORE_SIZE, body, size);
}

inline static void
subrecs_replace_min(void *subrecs, int size, int n_subrecs, int score, void *body, int dir)
{
  void *v;
  int n = 0, n1, n2, *c1, *c2;
  for (;;) {
    n1 = n * 2 + 1;
    n2 = n1 + 1;
    c1 = n1 < n_subrecs ? SUBRECS_NTH(subrecs,size,n1) : NULL;
    c2 = n2 < n_subrecs ? SUBRECS_NTH(subrecs,size,n2) : NULL;
    if (c1 && SUBRECS_CMP(score, *c1, dir)) {
      if (c2 && SUBRECS_CMP(score, *c2, dir) && SUBRECS_CMP(*c1, *c2, dir)) {
	SUBRECS_COPY(subrecs,size,n,c2);
	n = n2;
      } else {
	SUBRECS_COPY(subrecs,size,n,c1);
	n = n1;
      }
    } else {
      if (c2 && SUBRECS_CMP(score, *c2, dir)) {
	SUBRECS_COPY(subrecs,size,n,c2);
	n = n2;
      } else {
	break;
      }
    }
  }
  v = subrecs + n * (size + SCORE_SIZE);
  memcpy(v, &score, SCORE_SIZE);
  memcpy(v + SCORE_SIZE, body, size);
}

sen_rc 
sen_records_group(sen_records *r, int limit, sen_group_optarg *optarg)
{
  sen_set *g;
  recinfo *ri;
  sen_rec_unit unit;
  sen_set_cursor *c;
  const sen_recordh *rh;
  void *key, *ekey, *gkey = NULL;
  int funcp, rsize, ssize, unit_size, dir;
  if (!r || !r->records) { return sen_invalid_argument; }
  if (optarg) {
    unit = sen_rec_userdef;
    rsize = optarg->key_size;
    funcp = optarg->func ? 1 : 0;
    dir = (optarg->mode == sen_sort_ascending) ? -1 : 1;
  } else {
    unit = sen_rec_document;
    rsize = rec_unit_size(unit);
    funcp = 0;
    dir = 1;
  }
  if (funcp) {
    ssize = r->record_size;
    gkey = SEN_MALLOC(rsize ? rsize : 8192);
  } else {
    if (r->record_size <= rsize) { return sen_invalid_argument; }
    ssize = r->record_size - rsize;
  }
  unit_size = SCORE_SIZE + ssize;
  if (!(c = sen_set_cursor_open(r->records))) { return sen_memory_exhausted; }
  if (!(g = sen_set_open(rsize, SCORE_SIZE + sizeof(int) + unit_size * limit, 0))) {
    sen_set_cursor_close(c);
    return sen_memory_exhausted;
  }
  while ((rh = sen_set_cursor_next(c, &key, (void **)&ri))) {
    if (funcp) {
      if (optarg->func(r, rh, &gkey, optarg->func_arg)) { continue; }
      ekey = key;
    } else {
      gkey = key;
      ekey = key + rsize;
    }
    {
      recinfo *gri;
      if (sen_set_get(g, gkey, (void **)&gri)) {
	gri->score += ri->score;
	gri->n_subrecs += 1;
	if (limit) {
	  if (limit < gri->n_subrecs) {
	    if (SUBRECS_CMP(ri->score, *gri->subrecs, dir)) {
	      subrecs_replace_min(gri->subrecs, ssize, limit, ri->score, ekey, dir);
	    }
	  } else {
	    subrecs_push(gri->subrecs, ssize, gri->n_subrecs, ri->score, ekey, dir);
	  }
	}
      }
    }
  }
  sen_set_cursor_close(c);
  sen_set_close(r->records);
  if (funcp) { SEN_FREE(gkey); }
  r->records = g;
  r->subrec_unit = r->record_unit;
  r->record_unit = unit; //sen_rec_document;??
  r->subrec_size = ssize;
  r->record_size = rsize;
  r->max_n_subrecs = limit;
  sen_records_cursor_clear(r);
  return sen_success;
}

typedef struct {
  sen_id rid;
  uint32_t sid;
  uint32_t pos;
} posinfo;

const sen_recordh *
sen_records_at(sen_records *r, const void *key,
	       unsigned section, unsigned pos, sen_recinfo *ri)
{
  posinfo pi;
  if (!r || !r->keys) { return 0; }
  if (r->record_unit == sen_rec_userdef) {
    return sen_set_at(r->records, key, (void **) ri);
  } else {
    if (!(pi.rid = sen_sym_at(r->keys, key))) { return 0; }
    pi.sid = section;
    pi.pos = pos;
    return sen_set_at(r->records, &pi, (void **) ri);
  }
}

sen_rc
sen_record_info(sen_records *r, const sen_recordh *rh, sen_recinfo *ri)
{
  sen_rc rc;
  posinfo *pi;
  recinfo *_ri;
  if (!r || !rh || !ri) { return sen_invalid_argument; }
  rc = sen_set_element_info(r->records, rh, (void **)&pi, (void **)&_ri);
  if (rc) { return rc; }
  ri->userkey = (void *)pi;
  switch (r->record_unit) {
  case sen_rec_document :
    ri->key = sen_sym_key(r->keys, pi->rid);
    ri->section = 0;
    ri->pos = 0;
    break;
  case sen_rec_section :
    ri->key = sen_sym_key(r->keys, pi->rid);
    ri->section = pi->sid;
    ri->pos = 0;
    break;
  case sen_rec_position :
    ri->key = sen_sym_key(r->keys, pi->rid);
    ri->section = pi->sid;
    ri->pos = pi->pos;
    break;
  default :
    ri->key = NULL;
    ri->section = 0;
    ri->pos = 0;
    break;
  }
  ri->score = _ri->score;
  ri->n_subrecs = _ri->n_subrecs;
  return sen_success;
}

int
sen_record_subrec_info(sen_records *r, const sen_recordh *rh, sen_recinfo *ri,
		       int from, int limit)
{
  void *p;
  posinfo *pi;
  recinfo *_ri;
  int i, n_subrecs, unit_size = SCORE_SIZE + r->subrec_size;
  if (!r || !rh || !ri) { return 0; }
  if (sen_set_element_info(r->records, rh, (void **)&pi, (void **)&_ri)) { return -1; }
  n_subrecs = _ri->n_subrecs;
  if (_ri->n_subrecs > r->max_n_subrecs) { n_subrecs = r->max_n_subrecs; }
  if (from + limit > n_subrecs) { limit = n_subrecs - from; }
  if (limit < 0) { limit = 0; }
  p = ((void *)_ri->subrecs) + from * unit_size;
  for (i = 0; i < limit; i++, ri++) {
    ri->score = *((intptr_t *)p);
    ri->userkey = p + sizeof(intptr_t);
    switch (r->record_unit) {
    case sen_rec_document :
      ri->key = sen_sym_key(r->keys, pi->rid);
      if (r->subrec_unit != sen_rec_userdef) {
        ri->section = *((int *)(p + sizeof(intptr_t)));
      } else {
	ri->section = 0;
      }
      if (r->subrec_unit == sen_rec_position) {
	ri->pos = *((int *)(p + sizeof(intptr_t) + sizeof(intptr_t)));
      } else {
	ri->pos = 0;
      }
      break;
    case sen_rec_section :
      ri->key = sen_sym_key(r->keys, pi->rid);
      ri->section = pi->sid;
      if (r->subrec_unit == sen_rec_position) {
	ri->pos = *((int *)(p + sizeof(intptr_t)));
      } else {
	ri->pos = 0;
      }
      break;
    default : 
      {
	posinfo *spi = (posinfo *)(p + sizeof(intptr_t));
	switch (r->subrec_unit) {
	case sen_rec_document :
	  ri->key = sen_sym_key(r->keys, spi->rid);
	  ri->section = 0;
	  ri->pos = 0;
	  break;
	case sen_rec_section :
	  ri->key = sen_sym_key(r->keys, spi->rid);
	  ri->section = spi->sid;
	  ri->pos = 0;
	  break;
	case sen_rec_position :
	  ri->key = sen_sym_key(r->keys, spi->rid);
	  ri->section = spi->sid;
	  ri->pos = spi->pos;
	  break;
	default :
	  ri->key = NULL;
	  ri->section = 0;
	  ri->pos = 0;
	  break;
	}
      }
      break;
    }
    ri->n_subrecs = 0;
    p += unit_size;
  }
  return i;
}

/* select */

#define B31    0x80000000
#define B30_00 0x7fffffff
#define BIT30_00(x) (x & B30_00)

inline static void
res_add(sen_records *r, posinfo *pi, uint32_t score, sen_sel_operator op)
{
  recinfo *ri;
  sen_set_eh *eh = NULL;

  switch (op) {
  case sen_sel_or :
    eh = sen_set_get(r->records, pi, (void **)&ri);
    break;
  case sen_sel_and :
    if ((eh = sen_set_at(r->records, pi, (void **)&ri))) {
      ri->n_subrecs |= B31;
    }
    break;
  case sen_sel_but :
    if ((eh = sen_set_at(r->records, pi, (void **)&ri))) {
      sen_set_del(r->records, eh);
      eh = NULL;
    }
    break;
  case sen_sel_adjust :
    if ((eh = sen_set_at(r->records, pi, (void **)&ri))) {
      ri->score += score;
      eh = NULL;
    }
    break;
  }

  if (eh) {
    int limit = r->max_n_subrecs;
    ri->score += score;
    ri->n_subrecs += 1;
    if (limit) {
      int dir = 1;
      int ssize = r->subrec_size;
      int n_subrecs = BIT30_00(ri->n_subrecs);
      void *ekey = ((void *)pi) + r->record_size;
      if (limit < n_subrecs) {
	if (SUBRECS_CMP(score, *ri->subrecs, dir)) {
	  subrecs_replace_min(ri->subrecs, ssize, limit, score, ekey, dir);
	}
      } else {
	subrecs_push(ri->subrecs, ssize, n_subrecs, score, ekey, dir);
      }
    }
  }
}

struct _btr_node {
  struct _btr_node *car;
  struct _btr_node *cdr;
  token_info *ti;
};

typedef struct _btr_node btr_node;

typedef struct {
  int n;
  token_info *min;
  token_info *max;
  btr_node *root;
  btr_node *nodes;
} btr;

inline static void
bt_zap(btr *bt)
{
  bt->n = 0;
  bt->min = NULL;
  bt->max = NULL;
  bt->root = NULL;
}

inline static btr *
bt_open(int size)
{
  btr *bt = SEN_MALLOC(sizeof(btr));
  if (bt) {
    bt_zap(bt);
    if (!(bt->nodes = SEN_MALLOC(sizeof(btr_node) * size))) {
      SEN_FREE(bt);
      bt = NULL;
    }
  }
  return bt;
}

inline static void
bt_close(btr *bt)
{
  if (!bt) { return; }
  SEN_FREE(bt->nodes);
  SEN_FREE(bt);
}

inline static void
bt_push(btr *bt, token_info *ti)
{
  int pos = ti->pos, minp = 1, maxp = 1;
  btr_node *node, *new, **last;
  new = bt->nodes + bt->n++;
  new->ti = ti;
  new->car = NULL;
  new->cdr = NULL;
  for (last = &bt->root; (node = *last);) {
    if (pos < node->ti->pos) {
      last = &node->car;
      maxp = 0;
    } else {
      last = &node->cdr;
      minp = 0;
    }
  }
  *last = new;
  if (minp) { bt->min = ti; }
  if (maxp) { bt->max = ti; }
}

inline static void
bt_pop(btr *bt)
{
  btr_node *node, *min, **last;
  for (last = &bt->root; (min = *last) && min->car; last = &min->car) ;
  if (min) {
    int pos = min->ti->pos, minp = 1, maxp = 1;
    *last = min->cdr;
    min->cdr = NULL;
    for (last = &bt->root; (node = *last);) {
      if (pos < node->ti->pos) {
	last = &node->car;
	maxp = 0;
      } else {
	last = &node->cdr;
	minp = 0;
      }
    }
    *last = min;
    if (minp) { bt->min = min->ti; }
    if (maxp) { bt->max = min->ti; }
  }
}

typedef enum {
  sen_wv_none = 0,
  sen_wv_static,
  sen_wv_dynamic
} sen_wv_mode;

inline static int
get_weight(sen_records *r, sen_id rid, int sid,
	   sen_wv_mode wvm, sen_select_optarg *optarg)
{
  switch (wvm) {
  case sen_wv_none :
    return 1;
  case sen_wv_static :
    return sid < optarg->vector_size ? optarg->weight_vector[sid] : 0;
  case sen_wv_dynamic :
    return optarg->func(r, sen_sym_key(r->keys, rid), sid, optarg->func_arg);
  }
}

sen_rc
sen_index_similar_search(sen_index *i, const char *string,
			 sen_records *r, sen_sel_operator op, sen_select_optarg *optarg)
{
  int *w1, limit;
  sen_id tid, *tp;
  sen_rc rc = sen_success;
  sen_set *h = sen_set_open(sizeof(sen_id), sizeof(int), 0);
  sen_lex *lex = sen_lex_open(i->lexicon, string, 0);
  while (!lex->status) {
    if ((tid = sen_lex_next(lex))) {
      sen_set_get(h, &tid, (void **)&w1);
      (*w1)++;
    }
  }
  sen_lex_close(lex);
  {
    sen_set_cursor *c = sen_set_cursor_open(h);
    int maxsize = sen_sym_size(i->keys) * sizeof(int);
    while (sen_set_cursor_next(c, (void **) &tp, (void **) &w1)) {
      *w1 += maxsize / sen_inv_estimate_size(i->inv, *tp);
    }
    sen_set_cursor_close(c);
  }
  limit = optarg->similarity_threshold
    ? (optarg->similarity_threshold > h->n_entries
       ? h->n_entries
       : optarg->similarity_threshold)
    : (h->n_entries >> 3) + 1;
  {
    int j, w2, rep;
    sen_inv_cursor *c;
    sen_inv_posting *pos;
    sen_wv_mode wvm = sen_wv_none;
    sen_set_sort_optarg arg = {sen_sort_descending, NULL, (void *)sizeof(sen_id), NULL};
    sen_set_eh *eh, *sorted = sen_set_sort(h, limit, &arg);
    rep = (r->record_unit == sen_rec_position || r->subrec_unit == sen_rec_position);
    if (optarg->func) {
      wvm = sen_wv_dynamic;
    } else if (optarg->weight_vector && optarg->vector_size) {
      wvm = sen_wv_static;
    }
    for (j = 0, eh = sorted; j < limit; j++, eh++) {
      sen_set_element_info(h, eh, (void **) &tp, (void **) &w1);
      if (!*tp || !(c = sen_inv_cursor_open(i->inv, *tp))) {
	sen_log("cursor open failed (%d)", *tp);
	continue;
      }
      if (rep) {
	while (!sen_inv_cursor_next(c)) {
	  pos = c->post;
	  if ((w2 = get_weight(r, pos->rid, pos->sid, wvm, optarg))) {
	    while (!sen_inv_cursor_next_pos(c)) {
	      res_add(r, (posinfo *) pos, *w1 * w2 * (1 + pos->score), op);
	    }
	  }
	}
      } else {
	while (!sen_inv_cursor_next(c)) {
	  pos = c->post;
	  if ((w2 = get_weight(r, pos->rid, pos->sid, wvm, optarg))) {
	    res_add(r, (posinfo *) pos, *w1 * w2 * (pos->tf + pos->score), op);
	  }
	}
      }
      sen_inv_cursor_close(c);
    }
    SEN_FREE(sorted);
  }
  sen_set_close(h);
  if (op == sen_sel_and) {
    recinfo *ri;
    sen_set_eh *eh;
    sen_set_cursor *c = sen_set_cursor_open(r->records);
    while ((eh = sen_set_cursor_next(c, NULL, (void **) &ri))) {
      if ((ri->n_subrecs & B31)) {
	ri->n_subrecs &= B30_00;
      } else {
	sen_set_del(r->records, eh);
      }
    }
    sen_set_cursor_close(c);
  }
  sen_log("hits=%d", r->records ? r->records->n_entries : 0);
  r->keys = i->keys;
  sen_records_cursor_clear(r);
  return rc;
}

sen_rc
sen_index_select(sen_index *i, const char *string,
		 sen_records *r, sen_sel_operator op, sen_select_optarg *optarg)
{
  btr *bt = NULL;
  sen_rc rc = sen_success;
  int rep, orp, weight, max_interval = 0;
  token_info *ti, **tis, **tip, **tie;
  uint32_t n = 0, rid, sid, nrid, nsid;
  sen_sel_mode mode = sen_sel_exact;
  sen_wv_mode wvm = sen_wv_none;
  if (!r) { return sen_invalid_argument; }
  if (optarg) {
    mode = optarg->mode;
    if (optarg->func) {
      wvm = sen_wv_dynamic;
    } else if (optarg->weight_vector && optarg->vector_size) {
      wvm = sen_wv_static;
    }
  }
  if (mode == sen_sel_similar) {
    return sen_index_similar_search(i, string, r, op, optarg);
  }
  rep = (r->record_unit == sen_rec_position || r->subrec_unit == sen_rec_position);
  orp = (r->record_unit == sen_rec_position || op == sen_sel_or);
  if (!(tis = SEN_MALLOC(sizeof(token_info *) * strlen(string) * 2))) {
    return sen_memory_exhausted;
  }
  if ((rc = token_info_build(i, string, tis, &n, mode))) { goto exit; }
  sen_log("n=%d (%s)", n, string);
  if (!n) { goto exit; }
  if (mode == sen_sel_near) {
    if (!(bt = bt_open(n))) { rc = sen_memory_exhausted; goto exit; }
    max_interval = optarg->max_interval;
  }
  qsort(tis, n, sizeof(token_info *), token_compare);
  tie = tis + n;
  for (tip = tis; tip < tie; tip++) {
    ti = *tip;
    sen_log("o=%d n=%d s=%d r=%d", ti->offset, ti->ntoken, ti->size, ti->rid);
  }
  for (;;) {
    rid = (*tis)->rid;
    sid = (*tis)->sid;
    for (tip = tis + 1, nrid = rid, nsid = sid + 1; tip < tie; tip++) {
      ti = *tip;
      if (token_info_skip(ti, rid, sid)) { goto exit; }
      if (ti->rid != rid || ti->sid != sid) {
	nrid = ti->rid;
	nsid = ti->sid;
	break;
      }
    }
    weight = get_weight(r, rid, sid, wvm, optarg);
    if (tip == tie && weight) {
      posinfo pi = {rid, sid, 0};
      if (orp || sen_set_at(r->records, &pi, NULL)) {
	int count = 0, noccur = 0, pos = 0, score = 0, tscore = 0, min, max;

#define SKIP_OR_BREAK(pos) {\
  if (token_info_skip_pos(ti, rid, sid, pos)) { break; } \
  if (ti->rid != rid || ti->sid != sid) { \
    nrid = ti->rid; \
    nsid = ti->sid; \
    break; \
  } \
}
	if (mode == sen_sel_near) {
	  bt_zap(bt);
	  for (tip = tis; tip < tie; tip++) {
	    ti = *tip;
	    SKIP_OR_BREAK(pos);
	    bt_push(bt, ti);
	  }
	  if (tip == tie) {
	    for (;;) {
	      ti = bt->min; min = ti->pos; max = bt->max->pos;
	      if (max - min <= max_interval) {
		if (rep) { pi.pos = min; res_add(r, &pi, weight, op); }
		noccur++;
		SKIP_OR_BREAK(max);
	      } else {
		SKIP_OR_BREAK(max - max_interval);
	      }
	      bt_pop(bt);
	    }
	  }
	} else {
	  for (tip = tis; ; tip++) {
	    if (tip == tie) { tip = tis; }
	    ti = *tip;
	    SKIP_OR_BREAK(pos);
	    if (ti->pos == pos) {
	      score += ti->score; count++;
	    } else {
	      score = ti->score; count = 1; pos = ti->pos;
	    }
	    if (count == n) {
	      if (rep) { pi.pos = pos; res_add(r, &pi, (score + 1) * weight, op); }
	      tscore += score;
	      score = 0; count = 0; pos++;
	      noccur++;
	    }
	  }
	}
	if (noccur && !rep) { res_add(r, &pi, (noccur + tscore) * weight, op); }
      }
    }
    if (token_info_skip(*tis, nrid, nsid)) { goto exit; }
  }
exit :
  sen_log("exit");
  for (tip = tis; tip < tis + n; tip++) {
    if (*tip) { token_info_close(*tip); }
  }
  SEN_FREE(tis); 
  if (op == sen_sel_and) {
    recinfo *ri;
    sen_set_eh *eh;
    sen_set_cursor *c = sen_set_cursor_open(r->records);
    while ((eh = sen_set_cursor_next(c, NULL, (void **) &ri))) {
      if ((ri->n_subrecs & B31)) {
	ri->n_subrecs &= B30_00;
      } else {
	sen_set_del(r->records, eh);
      }
    }
    sen_set_cursor_close(c);
  }
  r->keys = i->keys;
  sen_records_cursor_clear(r);
  bt_close(bt);
  return rc;
}

sen_records *
sen_index_sel(sen_index *i, const char *string)
{
  sen_log("sen_index_sel > (%s)", string);
  {
  sen_select_optarg arg = {sen_sel_exact, 0, 0, NULL, 0, NULL, NULL};
  sen_records *r = sen_records_open(sen_rec_document, sen_rec_none, 0);
  sen_index_select(i, string, r, sen_sel_or, &arg);
  sen_log("exact: %d", r->records->n_entries);
  if (r->records->n_entries <= sen_partial_match_threshold) {
    arg.mode = sen_sel_partial;
    sen_index_select(i, string, r, sen_sel_or, &arg);
    sen_log("partial: %d", r->records->n_entries);
  }
  sen_log("hits=%d", r->records->n_entries);
  if (!r->records->n_entries) {
    sen_records_close(r);
    r = NULL;
  }
  return r;
  }
}
