/* Copyright(C) 2004 Brazil

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
  
  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/param.h>
#include "str.h"
#include "io.h"

#define SEN_IO_IDSTR "SENNA:IO:00.000"

#define FILE_SIZE               1073741824UL

static int page_size = 0;

struct _sen_io_header {
  char idstr[16];
  uint32_t header_size;
  uint32_t segment_size;
  uint32_t max_segment;
};

typedef struct {
  void *map;
  void **handle;
} mapinfo;

struct _sen_io {
  char path[PATH_MAX];
  struct _sen_io_header *header;
  void *user_header;
  mapinfo *map;
  uint32_t base;
  uint32_t mode;
  uint32_t cache_size;
  // pthread_mutex_t map_mutex;
};

sen_io *
sen_io_create(const char *path, int header_size, uint32_t segment_size,
	      uint32_t max_segment, int mode, int cache_size)
{
  int fd = -1, b;
  sen_io *io = NULL;
  struct _sen_io_header *header;
  if (!path || !*path || (strlen(path) > PATH_MAX - 4)) { goto exit; }
  if ((fd = open(path, O_RDWR|O_CREAT|O_TRUNC, 0666)) == -1) { goto exit; }
  if (!page_size) { page_size = getpagesize(); }
  for (b = page_size; b < sizeof(struct _sen_io_header) + header_size; b += page_size);
  if (ftruncate(fd, b) == -1) { goto exit; }
  header = SEN_MMAP(b, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);
  if (header == MAP_FAILED) { goto exit; }
  header->header_size = header_size;
  header->segment_size = segment_size;
  header->max_segment = max_segment;
  memcpy(header->idstr, SEN_IO_IDSTR, 16);
  msync(header, b, MS_SYNC);
  if (!(io = SEN_MALLOC(sizeof(sen_io)))) {
    SEN_MUNMAP(header, b);
    goto exit;
  }
  if (!(io->map = SEN_MALLOC(sizeof(mapinfo) * max_segment))) {
    SEN_FREE(io);
    io = NULL;
    SEN_MUNMAP(header, b);
    goto exit;
  }
  memset(io->map, 0, sizeof(mapinfo) * max_segment);
  strncpy(io->path, path, PATH_MAX);
  io->header = header;
  io->user_header = ((void *) header) + sizeof(struct _sen_io_header);
  io->base = b;
  io->mode = mode;
  io->cache_size = cache_size;
  // pthread_mutex_init(&io->map_mutex, NULL);
exit :
  if (fd != -1) { close(fd); }
  return io;
}

sen_io *
sen_io_open(const char *path, int mode, int cache_size)
{
  struct stat s;
  int fd = -1, b, header_size;
  sen_io *io = NULL;
  struct _sen_io_header *header;
  if (!path || !*path || (strlen(path) > PATH_MAX - 4)) { goto exit; }
  if ((fd = open(path, O_RDWR)) == -1) { goto exit; }
  if (!page_size) { page_size = getpagesize(); }
  if (fstat(fd, &s) == -1 || s.st_size < page_size) { goto exit; }
  header = SEN_MMAP(page_size, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);
  if (header == MAP_FAILED) { goto exit; }
  /*
  if (memcmp(header->idstr, SEN_IO_IDSTR, 16)) {
    SEN_MUNMAP(header, page_size);
    goto exit;
  }
  */
  header_size = header->header_size;
  for (b = page_size; b < sizeof(struct _sen_io_header) + header_size; b += page_size);
  if (b != page_size) {
    SEN_MUNMAP(header, page_size);
    if (s.st_size < b) { goto exit; }
    header = SEN_MMAP(b, PROT_READ|PROT_WRITE, MAP_SHARED, fd, 0);
    if (header == MAP_FAILED) { goto exit; }
  }
  if (!(io = SEN_MALLOC(sizeof(sen_io)))) {
    SEN_MUNMAP(header, b);
    goto exit;
  }
  if (!(io->map = SEN_MALLOC(sizeof(mapinfo) * header->max_segment))) {
    SEN_FREE(io);
    io = NULL;
    SEN_MUNMAP(header, b);
    goto exit;
  }
  memset(io->map, 0, sizeof(mapinfo) * header->max_segment);
  strncpy(io->path, path, PATH_MAX);
  io->header = header;
  io->user_header = ((void *) header) + sizeof(struct _sen_io_header);
  io->base = b;
  io->mode = mode;
  io->cache_size = cache_size;
exit :
  if (fd != -1) { close(fd); }
  return io;
}

sen_rc
sen_io_close(sen_io *io)
{
  int i;
  mapinfo *p;
  uint32_t segment_size = io->header->segment_size;
  for (p = io->map, i = io->header->max_segment; i; p++, i--) {
    if (p->map) {
      SEN_MUNMAP(p->map, segment_size);
      if (p->handle) { *p->handle = NULL; }
    }
  }
  // pthread_mutex_destroy(&io->map_mutex);
  SEN_MUNMAP(io->header, io->base);
  SEN_FREE(io->map);
  SEN_FREE(io);
  return sen_success;
}

void *
sen_io_header(sen_io *io)
{
  return io->user_header;
}

inline static void
itoh(unsigned int i, char *p, unsigned int len)
{
  static const char *hex = "0123456789ABCDEF";
  p += len;
  *p-- = '\0';
  while (len--) {
    *p-- = hex[i & 0xf];
    i >>= 4;
  }
}

void *
sen_io_win_map(sen_io *io, sen_io_win *iw,
	       uint32_t segment, uint32_t offset, uint32_t size, int mode)
{
  void *p;
  uint32_t nseg;
  uint32_t segment_size = io->header->segment_size;
  uint32_t segments_per_file = FILE_SIZE / segment_size;
  if (offset >= segment_size) {
    segment += offset / segment_size;
    offset = offset % segment_size;
  }
  nseg = (offset + size) / segment_size;
  if ((offset + size) % segment_size) { nseg++; }
  if (!size || !io || segment + nseg > io->header->max_segment ||
      (segment / segments_per_file != (segment + nseg - 1) / segments_per_file)) {
    return NULL;
  }
  switch (mode) {
  case sen_io_rdonly:
    {
      char path[PATH_MAX];
      int pos, fno = segment / segments_per_file;
      int fd, len = strlen(io->path);
      memcpy(path, io->path, len);
      if (fno) {
	path[len] = '.';
	itoh(fno, path + len + 1, 3);
	pos = (segment % segments_per_file) * segment_size + offset;
      } else {
	path[len] = '\0';
	pos = io->base + segment * segment_size + offset;
      }
      if (!(p = SEN_MALLOC(size))) { return NULL; }
      if ((fd = open(path, O_RDONLY)) == -1) { return NULL; }
      if (lseek(fd, pos, SEEK_SET) == -1 || read(fd, p, size) == -1) {
	close(fd);
	return NULL;
      }
      close(fd);
      iw->addr = p;
    }
    break;
  case sen_io_rdwr:
    // if (nseg > 1) { /* auto unmap is not implemented yet */
    if (nseg > 0) {
      struct stat s;
      char path[PATH_MAX];
      int pos, fno = segment / segments_per_file;
      int fd, len = strlen(io->path);
      uint32_t mapsize = nseg * segment_size;
      memcpy(path, io->path, len);
      if (fno) {
	path[len] = '.';
	itoh(fno, path + len + 1, 3);
	pos = (segment % segments_per_file) * segment_size;
      } else {
	path[len] = '\0';
	pos = io->base + segment * segment_size;
      }
      if ((fd = open(path, O_RDWR|O_CREAT, 0666)) == -1) {
	return NULL;
      }
      if ((fstat(fd, &s) == -1) ||
	  (s.st_size < pos + mapsize && ftruncate(fd, pos + mapsize) == -1)) {
	close(fd);
	return NULL;
      }
      p = SEN_MMAP(mapsize, PROT_READ|PROT_WRITE, MAP_SHARED, fd, pos);
      close(fd);
      if (p == MAP_FAILED) { return NULL; }
    } else {
      if (!(p = sen_io_seg_map(io, segment))) { return NULL; }
    }
    iw->addr = p + offset;
    break;
  case sen_io_wronly: 
    if (!(p = SEN_MALLOC(size))) { return NULL; }
    memset(p, 0, size);
    iw->addr = p;
    break;
  default :
    return NULL;
  }
  iw->io = io;
  iw->mode = mode;
  iw->segment = segment;
  iw->offset = offset;
  iw->nseg = nseg;
  iw->size = size;
  return iw->addr;
}

sen_rc
sen_io_win_unmap(sen_io_win *iw)
{
  sen_rc rc = sen_success;
  sen_io *io = iw->io;
  uint32_t segment_size = io->header->segment_size;
  uint32_t segments_per_file = FILE_SIZE / segment_size;
  int nseg = iw->nseg;
  switch (iw->mode) {
  case sen_io_rdonly:
    if (iw->addr) { SEN_FREE(iw->addr); }
    iw->addr = NULL;
    break;
  case sen_io_rdwr:
    // if (nseg > 1) { /* auto unmap is not implemented yet */
    if (nseg > 0) {
      SEN_MUNMAP(iw->addr - iw->offset, nseg * segment_size);
    } else {
      rc = sen_io_seg_unmap(io, iw->segment);
    }
    iw->addr = NULL;
    break;
  case sen_io_wronly: 
    {
      char path[PATH_MAX];
      int pos, fno = iw->segment / segments_per_file;
      int fd, len = strlen(io->path);
      memcpy(path, io->path, len);
      if (fno) {
	path[len] = '.';
	itoh(fno, path + len + 1, 3);
	pos = (iw->segment % segments_per_file) * segment_size + iw->offset;
      } else {
	path[len] = '\0';
	pos = io->base + iw->segment * segment_size + iw->offset;
      }
      if ((fd = open(path, O_WRONLY|O_CREAT, 0666)) == -1) {
	return sen_file_operation_error;
      }
      if (lseek(fd, pos, SEEK_SET) == -1 ||
	  write(fd, iw->addr, iw->size) == -1) {
	rc = sen_file_operation_error;
      } else {
	SEN_FREE(iw->addr);
	iw->addr = NULL;
      }
      close(fd);
    }
    break;
  default :
    rc = sen_invalid_argument;
  }
  return rc;
}

void * 
sen_io_seg_map(sen_io *io, int32_t segno)
{
  void *map;
  mapinfo *info;
  if (0 > segno || segno >= io->header->max_segment) { return NULL; }
  info = &io->map[segno];
  if (!(map = info->map)) {
    struct stat s;
    char path[PATH_MAX];
    uint32_t segment_size = io->header->segment_size;
    uint32_t segments_per_file = FILE_SIZE / segment_size;
    int pos, fno = segno / segments_per_file;
    int fd, len = strlen(io->path);
    memcpy(path, io->path, len);
    if (fno) {
      path[len] = '.';
      itoh(fno, path + len + 1, 3);
      pos = (segno % segments_per_file) * segment_size;
    } else {
      path[len] = '\0';
      pos = io->base + segno * segment_size;
    }
    if ((fd = open(path, O_RDWR|O_CREAT, 0666)) == -1) {
      return NULL;
    }
    if ((fstat(fd, &s) == -1) ||
	(s.st_size < pos + segment_size && ftruncate(fd, pos + segment_size) == -1)) {
      close(fd);
      return NULL;
    }
    // pthread_mutex_lock(&io->map_mutex);
    if (!(map = info->map)) {
      map = SEN_MMAP(segment_size, PROT_READ|PROT_WRITE, MAP_SHARED, fd, pos);
      if (map == MAP_FAILED) {
	map = NULL;
      } else {
	info->map = map;
      }
    }
    // pthread_mutex_unlock(&io->map_mutex);
    close(fd);
    //    sen_log("seg_map segno=%d map=%p", segno, map);
  }
  return map;
}

sen_rc
sen_io_seg_unmap(sen_io *io, int32_t segno)
{
  void *map;
  mapinfo *info;
  if (0 > segno || segno >= io->header->max_segment) { return sen_invalid_argument; }
  info = &io->map[segno];
  // mutex_lock();
  if ((map = info->map)) {
    // todo:
    // enqueue(io->garbases, segno);
    //info->map = NULL;
    if (info->handle) { *(info->handle) = NULL; }
  }
  // mutex_unlock();
  return sen_success;
}

sen_rc
sen_io_seg_bind(sen_io *io, sen_io_seginfo *si)
{
  if ((si->addr = sen_io_seg_map(io, si->segno))) {
    io->map[si->segno].handle = &si->addr;
    return sen_success;
  } else {
    return sen_memory_exhausted;
  }
}

sen_rc
sen_io_seg_unbind(sen_io *io, sen_io_seginfo *si)
{
  sen_rc rc;
  if (io->map[si->segno].handle != &si->addr) {
    sen_log("invalid unbind request");
    return sen_invalid_argument;
  }
  rc = sen_io_seg_unmap(io, si->segno);
  si->segno = -1;
  si->addr = NULL;
  return rc;
}

// todo: sen_io_gc()

/*
enque(io)
{
  pickup unused segments using mincore;
  enqueue;
}

unmap(io)
{
  unmap segments in queue and passed enough time;
}
*/
