/* Copyright(C) 2004 Brazil

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
  
  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <string.h>
#include <sys/mman.h>
#include "senna_in.h"
#include "str.h"
#include "set.h"

static sen_set *prefix = NULL;
static sen_set *suffix = NULL;

#define N_PREFIX 2048
#define N_SUFFIX 0

#define PREFIX_PATH SENNA_HOME "/prefix"
#define SUFFIX_PATH SENNA_HOME "/suffix"

inline static void
prefix_init(void)
{
  int i, *ip;
  FILE *fp;
  char buffer[4];
  prefix = sen_set_open(2, sizeof(int), 0);
  if ((fp = fopen(PREFIX_PATH, "r"))) {
    for (i = 0; i < N_PREFIX; i++) {
      if (!fgets(buffer, 4, fp)) { break; }
      sen_set_get(prefix, buffer, (void **)&ip);
      *ip = i;
    }
    fclose(fp);
  }
}

inline static void
suffix_init(void)
{
  int i;
  FILE *fp;
  char buffer[4];
  suffix = sen_set_open(2, 0, 0);
  if ((fp = fopen(SUFFIX_PATH, "r"))) {
    for (i = N_SUFFIX; i; i--) {
      if (!fgets(buffer, 4, fp)) { break; }
      sen_set_get(suffix, buffer, NULL);
    }
    fclose(fp);
  }
}

sen_rc
sen_str_fin(void)
{
  if (prefix) { sen_set_close(prefix); }
  if (suffix) { sen_set_close(suffix); }
  return sen_success;
}

int
sen_str_get_prefix_order(const char *str)
{
  int *ip;
  if (!prefix) { prefix_init(); }
  if (sen_set_at(prefix, str, (void **)&ip)) {
    return *ip;
  } else {
    return -1;
  }
}

inline static int
char_type(unsigned char *str, int *length)
{
  unsigned char c = *str;

  if (!suffix) { suffix_init(); }

  if (c & 0x80) {
    if (!str[1]) {
      *length = 1;
      return sen_str_others;
    }
    *length = 2;

    if (sen_set_at(suffix, str, NULL)) {
      return sen_str_symbol;
    }

    switch (c >> 4) {
    case 0x08 :
      return (c == 0x8e) ? sen_str_katakana : sen_str_others;
    case 0x09 :
      return sen_str_others;
    case 0x0a :
      switch (c & 0x0f) {
      case 1 :
	switch (str[1]) {
	case 0xbc :
	  return sen_str_katakana;
	case 0xb9 :
	  return sen_str_kanji;
	default :
	  return sen_str_symbol;
	}
      case 2 :
	return sen_str_symbol;
      case 3 :
	return (str[1] <= 0x39) ? sen_str_digit : sen_str_alpha;
      case 4 :
	return sen_str_hiragana;
      case 5 :
	return sen_str_katakana;
      case 6 :
      case 7 :
      case 8 :
	return sen_str_symbol;
      default :
	return sen_str_others;
      }
    default :
      return sen_str_kanji;
    }
  } else {
    *length = 1;
    switch (c >> 4) {
    case 2 :
      return sen_str_symbol;
    case 3 :
      return (c <= 0x39) ? sen_str_digit : sen_str_symbol;
    case 4 :
      return (c == 0x40) ? sen_str_symbol : sen_str_alpha;
    case 5 :
      return (c <= 0x5a) ? sen_str_alpha : sen_str_symbol;
    case 6 :
      return (c == 0x60) ? sen_str_symbol : sen_str_alpha;
    case 7 :
      return (c <= 0x7a) ? sen_str_alpha : (c == 0x7f ? sen_str_others : sen_str_symbol);
    default :
      return sen_str_others;
    }
  }
}

inline static char *
get_substr(char *str, sen_substr *res)
{
  int len, ctype = char_type(str, &len);
  if (ctype == sen_str_symbol) {
    res->ctype = ctype;
    res->start = str;
    res->length = len;
    return str + len;
  } else {
    char *p;
    for (p = str + len; *p; p += len) {
      if (char_type(p, &len) != ctype) { break; }
    }
    res->ctype = ctype;
    res->start = str;
    res->length = p - str;
    return p;
  }
}

int
sen_str_split(char *str, sen_substr *res, int max, char **rest)
{
  int i;
  sen_substr *r;
  char *p = str;
  if (!p) { return 0; }
  for (r = res, i = 0; i < max; r++, i++) {
    if (!*p) { break; }
    p = get_substr(p, r);
  }
  if (rest) { *rest = p; }
  return i;
}

struct _sen_str_splitter {
  char *next;
  char backup; 
};

sen_str_splitter *
sen_str_splitter_open(char *str)
{
  sen_str_splitter *s;
  if (!(s = SEN_MALLOC(sizeof(sen_str_splitter)))) { return NULL; }
  s->next = str;
  s->backup = *str;
  return s;
}

char *
sen_str_splitter_next(sen_str_splitter *s, int *ctype)
{
  int c, l;
  char *p, *q;
  p = s->next;
  *p = s->backup;
  if (!*p) { return NULL; }
  c = char_type(p, &l);
  q = p + l;
  if (c != sen_str_symbol) {
    for (; *q; q += l) {
      if (char_type(q, &l) != c) { break; }
    }
  }
  s->next = q;
  s->backup = *q;
  *q = '\0';
  if (ctype) { *ctype = c; }
  return p;
}

sen_rc
sen_str_splitter_close(sen_str_splitter *s)
{
  if (!s) { return sen_invalid_argument; }
  *s->next = s->backup;
  SEN_FREE(s);
  return sen_success;
}

#define BUFSIZE 65536

#include <stdio.h>
#include <stdarg.h>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>

void
sen_log(char *fmt, ...)
{
  static FILE *fp = NULL;
  // static unsigned int count = 0;
  va_list argp;
  char buffer[BUFSIZE];
  va_start(argp, fmt);
  vsnprintf(buffer, BUFSIZE - 1, fmt, argp);
  va_end(argp);
  buffer[BUFSIZE - 1] = '\0';
  if (!fp) { fp = fopen(SENNA_HOME "/log/senna.log", "a"); }
  if (fp) {
    struct timeval tv;
    struct tm tm;
    gettimeofday(&tv, NULL);
    localtime_r((time_t *)&tv.tv_sec, &tm);
    fprintf(fp, "%02d/%02d:%02d:%02d:%02d.%06d|%04x|%s\n",
	    tm.tm_mon + 1, tm.tm_mday,
	    tm.tm_hour, tm.tm_min, tm.tm_sec, (int) tv.tv_usec, getpid(), buffer);
    fflush(fp);
    /*
    if (!(++count % 1024)) {
      fclose(fp);
      fp = NULL;
    }
    */
  }
}

// static int alloc_size = 0;
static int alloc_count = 0;
static int mmap_size = 0;

void *
sen_malloc(size_t size, const char* file, int line)
{
  void *res;
  // res = malloc(size);
  res = calloc(1, size);
  if (res) { alloc_count++; }
  if (!res) { sen_log("malloc(%d)=%p (%s:%d) <%d>", size, res, file, line, alloc_count); }
  return res;
}

void
sen_free(void *ptr, const char* file, int line)
{
  free(ptr);
  if (ptr) { alloc_count--; }
  if (!ptr) { sen_log("free(%p) (%s:%d) <%d>", ptr, file, line, alloc_count); }
}

void *
sen_realloc(void *ptr, size_t size, const char* file, int line)
{
  void *res;
  res = realloc(ptr, size);
  if (res) { alloc_count++; }
  sen_log("realloc(%p,%d)=%p (%s:%d) <%d>", ptr, size, res, file, line, alloc_count);
  return res;
}

char *
sen_strdup(const char *s, const char* file, int line)
{
  char *res;
  res = strdup(s);
  if (res) { alloc_count++; }
  sen_log("strdup(%p)=%p (%s:%d) <%d>", s, res, file, line, alloc_count);
  return res;
}

void *
sen_mmap(size_t length, int prot, int flags, int fd, off_t offset,
	 const char* file, int line)
{
  void *res;
  res = mmap(NULL, length, prot, flags, fd, offset);
  mmap_size += length;
  if (MAP_FAILED == res) {
    sen_log("mmap(%d,%d,%d)=%p (%s:%d) <%d>", length, fd, offset, res, file, line, mmap_size);
  }
  return res;
}

int
sen_munmap(void *start, size_t length, const char* file, int line)
{
  int res;
  res = munmap(start, length);
  mmap_size -= length;
  if (res) { sen_log("munmap(%p,%d)=%d (%s:%d) <%d>", start, length, res, file, line, mmap_size); }
  return res;
}

static unsigned char symbol[] = {
  ',', '.', 0, ':', ';', '?', '!', 0, 0, 0, '`', 0, '^', '~', '_', 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, '-', '-', '/', '\\', 0, 0, '|', 0, 0, 0, '\'', 0,
  '"', '(', ')', 0, 0, '[', ']', '{', '}', 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  '+', '-', 0, 0, 0, '=', 0, '<', '>', 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  '$', 0, 0, '%', '#', '&', '*', '@', 0, 0, 0, 0, 0, 0, 0, 0
};

inline static int
normalize_euc(const char *str, char *norm, uint_least8_t *ctypes, int removeblankp)
{
  static uint16_t hankana[] = {
    0xa1a1, 0xa1a3, 0xa1d6, 0xa1d7, 0xa1a2, 0xa1a6, 0xa5f2, 0xa5a1, 0xa5a3,
    0xa5a5, 0xa5a7, 0xa5a9, 0xa5e3, 0xa5e5, 0xa5e7, 0xa5c3, 0xa1bc, 0xa5a2,
    0xa5a4, 0xa5a6, 0xa5a8, 0xa5aa, 0xa5ab, 0xa5ad, 0xa5af, 0xa5b1, 0xa5b3,
    0xa5b5, 0xa5b7, 0xa5b9, 0xa5bb, 0xa5bd, 0xa5bf, 0xa5c1, 0xa5c4, 0xa5c6,
    0xa5c8, 0xa5ca, 0xa5cb, 0xa5cc, 0xa5cd, 0xa5ce, 0xa5cf, 0xa5d2, 0xa5d5,
    0xa5d8, 0xa5db, 0xa5de, 0xa5df, 0xa5e0, 0xa5e1, 0xa5e2, 0xa5e4, 0xa5e6,
    0xa5e8, 0xa5e9, 0xa5ea, 0xa5eb, 0xa5ec, 0xa5ed, 0xa5ef, 0xa5f3, 0xa1ab,
    0xa1eb
  };
  static unsigned char dakuten[] = {
    0xf4, 0, 0, 0, 0, 0xac, 0, 0xae, 0, 0xb0, 0, 0xb2, 0, 0xb4, 0, 0xb6, 0,
    0xb8, 0, 0xba, 0, 0xbc, 0, 0xbe, 0, 0xc0, 0, 0xc2, 0, 0, 0xc5, 0, 0xc7,
    0, 0xc9, 0, 0, 0, 0, 0, 0, 0xd0, 0, 0, 0xd3, 0, 0, 0xd6, 0, 0, 0xd9, 0,
    0, 0xdc
  };
  static unsigned char handaku[] = {
    0xd1, 0, 0, 0xd4, 0, 0, 0xd7, 0, 0, 0xda, 0, 0, 0xdd
  };
  uint_least8_t *cp = ctypes, ctype;
  const unsigned char *s;
  unsigned char *d, *d0 = (unsigned char *) norm, b;
  for (s = (unsigned char *) str, d = d0; *s; s++) {
    if ((*s & 0x80)) {
      if ((*(s + 1) & 0x80)) {
	unsigned char c1 = *s++, c2 = *s, c3 = 0;
	switch (c1 >> 4) {
	case 0x08 :
	  if (c1 == 0x8e && 0xa0 <= c2 && c2 <= 0xdf) {
	    uint16_t c = hankana[c2 - 0xa0];
	    switch (c) {
	    case 0xa1ab :
	      if (d > d0 + 1 && d[-2] == 0xa5
		  && 0xa6 <= d[-1] && d[-1] <= 0xdb && (b = dakuten[d[-1] - 0xa6])) {
		*(d - 1) = b;
		continue;
	      } else {
		*d++ = c >> 8; *d = c & 0xff;
	      }
	      break;
	    case 0xa1eb :
	      if (d > d0 + 1 && d[-2] == 0xa5
		  && 0xcf <= d[-1] && d[-1] <= 0xdb && (b = handaku[d[-1] - 0xcf])) {
		*(d - 1) = b;
		continue;
	      } else {
		*d++ = c >> 8; *d = c & 0xff;
	      }
	      break;
	    default :
	      *d++ = c >> 8; *d = c & 0xff;
	      break;
	    }
	    ctype = sen_str_katakana;
	  } else {
	    *d++ = c1; *d = c2;
	    ctype = sen_str_others;
	  }
	  break;
	case 0x09 :
	  *d++ = c1; *d = c2;
	  ctype = sen_str_others;
	  break;
	case 0x0a :
	  switch (c1 & 0x0f) {
	  case 1 :
	    switch (c2) {
	    case 0xbc :
	      *d++ = c1; *d = c2;
	      ctype = sen_str_katakana;
	      break;
	    case 0xb9 :
	      *d++ = c1; *d = c2;
	      ctype = sen_str_kanji;
	      break;
	    case 0xa1 :
	      if (removeblankp) {
		if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
		continue;
	      } else {
		*d = ' ';
		ctype = SEN_NSTR_BLANK|sen_str_symbol;
	      }
	      break;
	    default :
	      if (c2 >= 0xa4 && (c3 = symbol[c2 - 0xa4])) {
		*d = c3;
		ctype = sen_str_symbol;
	      } else {
		*d++ = c1; *d = c2;
		ctype = sen_str_others;
	      }
	      break;
	    }
	    break;
	  case 2 :
	    *d++ = c1; *d = c2;
	    ctype = sen_str_symbol;
	    break;
	  case 3 :
	    c3 = c2 - 0x80;
	    if ('a' <= c3 && c3 <= 'z') {
	      ctype = sen_str_alpha;
	      *d = c3;
	    } else if ('A' <= c3 && c3 <= 'Z') {
	      ctype = sen_str_alpha;
	      *d = c3 + 0x20;
	    } else if ('0' <= c3 && c3 <= '9') {
	      ctype = sen_str_digit;
	      *d = c3;
	    } else {
	      ctype = sen_str_others;
	      *d++ = c1; *d = c2;
	    }
	    break;
	  case 4 :
	    *d++ = c1; *d = c2;
	    ctype = sen_str_hiragana;
	    break;
	  case 5 :
	    *d++ = c1; *d = c2;
	    ctype = sen_str_katakana;
	    break;
	  case 6 :
	  case 7 :
	  case 8 :
	    *d++ = c1; *d = c2;
	    ctype = sen_str_symbol;
	    break;
	  default :
	    *d++ = c1; *d = c2;
	    ctype = sen_str_others;
	    break;
	  }
	  break;
	default :
	  *d++ = c1; *d = c2;
	  ctype = sen_str_kanji;
	  break;
	}
      } else {
	/* skip invalid character */
	continue;
      }
    } else {
      unsigned char c = *s;
      switch (c >> 4) {
      case 0 :
      case 1 :
	/* skip unprintable ascii */
	if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
	continue;
      case 2 :
	if (c == 0x20) {
	  if (removeblankp) {
	    if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
	    continue;
	  } else {
	    *d = ' ';
	    ctype = SEN_NSTR_BLANK|sen_str_symbol;
	  }
	} else {
	  *d = c;
	  ctype = sen_str_symbol;
	}
	break;
      case 3 :
	*d = c;
	ctype = (c <= 0x39) ? sen_str_digit : sen_str_symbol;
	break;
      case 4 :
	*d = ('A' <= c) ? c + 0x20 : c;
	ctype = (c == 0x40) ? sen_str_symbol : sen_str_alpha;
	break;
      case 5 :
	*d = (c <= 'Z') ? c + 0x20 : c;
	ctype = (c <= 0x5a) ? sen_str_alpha : sen_str_symbol;
	break;
      case 6 :
	*d = c;
	ctype = (c == 0x60) ? sen_str_symbol : sen_str_alpha;
	break;
      case 7 :
	*d = c;
	ctype = (c <= 0x7a) ? sen_str_alpha : (c == 0x7f ? sen_str_others : sen_str_symbol);
	break;
      default :
	*d = c;
	ctype = sen_str_others;
	break;
      }
    }
    *cp++ = ctype;
    d++;
  }
  *cp = sen_str_null;
  *d = '\0';
  return cp - ctypes;
}

inline static int
normalize_utf8(const char *str, char *norm, uint_least8_t *ctypes, int removeblankp)
{
  // todo;
  return 0;
}

inline static int
normalize_sjis(const char *str, char *norm, uint_least8_t *ctypes, int removeblankp)
{
  static uint16_t hankana[] = {
    0x8140, 0x8142, 0x8175, 0x8176, 0x8141, 0x8145, 0x8392, 0x8340, 0x8342,
    0x8344, 0x8346, 0x8348, 0x8383, 0x8385, 0x8387, 0x8362, 0x815b, 0x8341,
    0x8343, 0x8345, 0x8347, 0x8349, 0x834a, 0x834c, 0x834e, 0x8350, 0x8352,
    0x8354, 0x8356, 0x8358, 0x835a, 0x835c, 0x835e, 0x8360, 0x8363, 0x8365,
    0x8367, 0x8369, 0x836a, 0x836b, 0x836c, 0x836d, 0x836e, 0x8371, 0x8374,
    0x8377, 0x837a, 0x837d, 0x837e, 0x8380, 0x8381, 0x8382, 0x8384, 0x8386,
    0x8388, 0x8389, 0x838a, 0x838b, 0x838c, 0x838d, 0x838f, 0x8393, 0x814a,
    0x814b
  };
  static unsigned char dakuten[] = {
    0x94, 0, 0, 0, 0, 0x4b, 0, 0x4d, 0, 0x4f, 0, 0x51, 0, 0x53, 0, 0x55, 0,
    0x57, 0, 0x59, 0, 0x5b, 0, 0x5d, 0, 0x5f, 0, 0x61, 0, 0, 0x64, 0, 0x66,
    0, 0x68, 0, 0, 0, 0, 0, 0, 0x6f, 0, 0, 0x72, 0, 0, 0x75, 0, 0, 0x78, 0,
    0, 0x7b
  };
  static unsigned char handaku[] = {
    0x70, 0, 0, 0x73, 0, 0, 0x76, 0, 0, 0x79, 0, 0, 0x7c
  };
  uint_least8_t *cp = ctypes, ctype;
  const unsigned char *s;
  unsigned char *d, *d0 = (unsigned char *) norm, b;
  for (s = (unsigned char *) str, d = d0; *s; s++) {
    if ((*s & 0x80)) {
      if (0xa0 <= *s && *s <= 0xdf) {
	uint16_t c = hankana[*s - 0xa0];
	switch (c) {
	case 0x814a :
	  if (d > d0 + 1 && d[-2] == 0x83
	      && 0x45 <= d[-1] && d[-1] <= 0x7a && (b = dakuten[d[-1] - 0x45])) {
	    *(d - 1) = b;
	    continue;
	  } else {
	    *d++ = c >> 8; *d = c & 0xff;
	  }
	  break;
	case 0x814b :
	  if (d > d0 + 1 && d[-2] == 0x83
	      && 0x6e <= d[-1] && d[-1] <= 0x7a && (b = handaku[d[-1] - 0x6e])) {
	    *(d - 1) = b;
	    continue;
	  } else {
	    *d++ = c >> 8; *d = c & 0xff;
	  }
	  break;
	default :
	  *d++ = c >> 8; *d = c & 0xff;
	  break;
	}
	ctype = sen_str_katakana;
      } else {
	if (0x40 <= *(s + 1) && *(s + 1) <= 0xfc) {
	  unsigned char c1 = *s++, c2 = *s, c3 = 0;
	  if (0x81 <= c1 && c1 <= 0x87) {
	    switch (c1 & 0x0f) {
	    case 1 :
	      switch (c2) {
	      case 0x5b :
		*d++ = c1; *d = c2;
		ctype = sen_str_katakana;
		break;
	      case 0x58 :
		*d++ = c1; *d = c2;
		ctype = sen_str_kanji;
		break;
	      case 0x40 :
		if (removeblankp) {
		  if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
		  continue;
		} else {
		  *d = ' ';
		  ctype = SEN_NSTR_BLANK|sen_str_symbol;
		}
		break;
	      default :
		if (0x43 <= c2 && c2 <= 0x7e && (c3 = symbol[c2 - 0x43])) {
		  *d = c3;
		  ctype = sen_str_symbol;
		} else if (0x7f <= c2 && c2 <= 0x97 && (c3 = symbol[c2 - 0x44])) {
		  *d = c3;
		  ctype = sen_str_symbol;
		} else {
		  *d++ = c1; *d = c2;
		  ctype = sen_str_others;
		}
		break;
	      }
	      break;
	    case 2 :
	      c3 = c2 - 0x1f;
	      if (0x4f <= c2 && c2 <= 0x58) {
		ctype = sen_str_digit;
		*d = c2 - 0x1f;
	      } else if (0x60 <= c2 && c2 <= 0x79) {
		ctype = sen_str_alpha;
		*d = c2 + 0x01;
	      } else if (0x81 <= c2 && c2 <= 0x9a) {
		ctype = sen_str_alpha;
		*d = c2 - 0x20;
	      } else if (0x9f <= c2 && c2 <= 0xf1) {
		*d++ = c1; *d = c2;
		ctype = sen_str_hiragana;
	      } else {
		*d++ = c1; *d = c2;
		ctype = sen_str_others;
	      }
	      break;
	    case 3 :
	      if (0x40 <= c2 && c2 <= 0x96) {
		*d++ = c1; *d = c2;
		ctype = sen_str_katakana;
	      } else {
		*d++ = c1; *d = c2;
		ctype = sen_str_symbol;
	      }
	      break;
	    case 4 :
	    case 7 :
	      *d++ = c1; *d = c2;
	      ctype = sen_str_symbol;
	      break;
	    default :
	      *d++ = c1; *d = c2;
	      ctype = sen_str_others;
	      break;
	    }
	  } else {
	    *d++ = c1; *d = c2;
	    ctype = sen_str_kanji;
	  }
	} else {
	  /* skip invalid character */
	  continue;
	}
      }
    } else {
      unsigned char c = *s;
      switch (c >> 4) {
      case 0 :
      case 1 :
	/* skip unprintable ascii */
	if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
	continue;
      case 2 :
	if (c == 0x20) {
	  if (removeblankp) {
	    if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
	    continue;
	  } else {
	    *d = ' ';
	    ctype = SEN_NSTR_BLANK|sen_str_symbol;
	  }
	} else {
	  *d = c;
	  ctype = sen_str_symbol;
	}
	break;
      case 3 :
	*d = c;
	ctype = (c <= 0x39) ? sen_str_digit : sen_str_symbol;
	break;
      case 4 :
	*d = ('A' <= c) ? c + 0x20 : c;
	ctype = (c == 0x40) ? sen_str_symbol : sen_str_alpha;
	break;
      case 5 :
	*d = (c <= 'Z') ? c + 0x20 : c;
	ctype = (c <= 0x5a) ? sen_str_alpha : sen_str_symbol;
	break;
      case 6 :
	*d = c;
	ctype = (c == 0x60) ? sen_str_symbol : sen_str_alpha;
	break;
      case 7 :
	*d = c;
	ctype = (c <= 0x7a) ? sen_str_alpha : (c == 0x7f ? sen_str_others : sen_str_symbol);
	break;
      default :
	*d = c;
	ctype = sen_str_others;
	break;
      }
    }
    *cp++ = ctype;
    d++;
  }
  *cp = sen_str_null;
  *d = '\0';
  return cp - ctypes;
}

inline static int
normalize_none(const char *str, char *norm, uint_least8_t *ctypes, int removeblankp)
{
  uint_least8_t *cp = ctypes, ctype;
  const unsigned char *s;
  unsigned char *d, *d0 = (unsigned char *) norm;
  for (s = (unsigned char *) str, d = d0; *s; s++) {
    unsigned char c = *s;
    switch (c >> 4) {
    case 0 :
    case 1 :
      /* skip unprintable ascii */
      if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
      continue;
    case 2 :
      if (c == 0x20) {
	if (removeblankp) {
	  if (cp > ctypes) { *(cp - 1) |= SEN_NSTR_BLANK; }
	  continue;
	} else {
	  *d = ' ';
	  ctype = SEN_NSTR_BLANK|sen_str_symbol;
	}
      } else {
	*d = c;
	ctype = sen_str_symbol;
      }
      break;
    case 3 :
      *d = c;
      ctype = (c <= 0x39) ? sen_str_digit : sen_str_symbol;
      break;
    case 4 :
      *d = ('A' <= c) ? c + 0x20 : c;
      ctype = (c == 0x40) ? sen_str_symbol : sen_str_alpha;
      break;
    case 5 :
      *d = (c <= 'Z') ? c + 0x20 : c;
      ctype = (c <= 0x5a) ? sen_str_alpha : sen_str_symbol;
      break;
    case 6 :
      *d = c;
      ctype = (c == 0x60) ? sen_str_symbol : sen_str_alpha;
      break;
    case 7 :
      *d = c;
      ctype = (c <= 0x7a) ? sen_str_alpha : (c == 0x7f ? sen_str_others : sen_str_symbol);
      break;
    default :
      *d = c;
      ctype = sen_str_others;
      break;
    }
    *cp++ = ctype;
    d++;
  }
  *cp = sen_str_null;
  *d = '\0';
  return cp - ctypes;
}

sen_nstr *
sen_nstr_open(const char *str, sen_encoding encoding, int removeblankp)
{
  size_t size;
  sen_nstr *r;
  if (!str || !(size = strlen(str))) { return NULL; }
  if (!(r = SEN_MALLOC(sizeof(sen_nstr)))) { return NULL; }
  r->orig = str;
  if (!(r->norm = SEN_MALLOC(size * 2 + 1))) {
    SEN_FREE(r);
    return NULL;
  }
  if (!(r->ctypes = SEN_MALLOC(size + 1))) {
    SEN_FREE(r->norm);
    SEN_FREE(r);
    return NULL;
  }
  switch (encoding) {
  case sen_enc_euc_jp :
    r->length = normalize_euc(str, r->norm, r->ctypes, removeblankp);
    break;
  case sen_enc_utf8 :
    r->length = normalize_utf8(str, r->norm, r->ctypes, removeblankp);
    break;
  case sen_enc_sjis :
    r->length = normalize_sjis(str, r->norm, r->ctypes, removeblankp);
    break;
  default :
    r->length = normalize_none(str, r->norm, r->ctypes, removeblankp);
    break;
  }
  return r;
}

sen_rc
sen_nstr_close(sen_nstr *nstr)
{
  if (!nstr) { return sen_invalid_argument; }
  SEN_FREE(nstr->norm);
  SEN_FREE(nstr->ctypes);
  SEN_FREE(nstr);
  return sen_success;
}

const char * sen_enc_string[] = {
  "default",
  "none",
  "euc_jp",
  "utf8",
  "sjis"
};

size_t
sen_str_charlen(const char *str, sen_encoding encoding)
{
  size_t size;
  const unsigned char *p = (const unsigned char *) str;
  if (!*p) { return 0; }
  switch (encoding) {
  case sen_enc_euc_jp :
    size = ((*p & 0x80) && (*(p + 1))) ? 2 : 1;
    break;
  case sen_enc_utf8 :
    if (*p & 0x80) {
      int b, w; 
      for (b = 0x40, w = 0; b && (*p & b); b >>= 1, w++);
      if (!w) { sen_log("invalid utf8 string"); }
      for (size = 1; w--; size++) {
	if (!*++p || (*p & 0xc0) != 0x80) {
	  sen_log("invalid utf8 string");
	  break;
	}
      }
    } else {
      size = 1;
    }
    break;
  case sen_enc_sjis :
    if (*p & 0x80) {
      if ((0xa1 <= *p && *p <= 0xdf) || !(*(p + 1))) {
	size = 1;
      } else {
	size = 2;
      }
    } else {
      size = 1;
    }
    break;
  default :
    size = 1;
    break;
  }
  return size;
}

size_t
sen_str_len(const char *str, sen_encoding encoding, const char **last)
{
  size_t len;
  const char *p = NULL;
  for (len = 0; *str; len++) {
    p = str;
    str += sen_str_charlen(str, encoding);
  }
  if (last) { *last = p; }
  return len;
}
