/* Copyright(C) 2004 Brazil

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
  
  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <fcntl.h>
#include <stdio.h> /* for debug */
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include "sym.h"

#define SEN_SYM_HEADER_SIZE 0x10000
#define SEN_IDSTR "SENNA:SYM:00.00"
#define SEN_SYM_MAX_POS 2147483647L
#define SEN_SYM_DELETED (SEN_SYM_MAX_ID + 1)

#define W_OF_KEY_IN_A_SEGMENT 22
#define W_OF_PAT_IN_A_SEGMENT 18
#define W_OF_SIS_IN_A_SEGMENT 19
#define KEY_MASK_IN_A_SEGMENT 0x3fffff
#define PAT_MASK_IN_A_SEGMENT 0x3ffff
#define SIS_MASK_IN_A_SEGMENT 0x7ffff

typedef struct {
  sen_id r;
  sen_id l;
  uint32_t key;
  uint16_t check;
  uint8_t deleting : 1;
  uint8_t immediate : 1;
  uint16_t pocket : 14;
} pat_node;

typedef struct {
  sen_id parent;
  sen_id sibling;
} sis_node;

enum {
  delinfo_empty = 0,
  delinfo_phase1,
  delinfo_phase2,
};

enum {
  segment_empty = 0,
  segment_key,
  segment_pat,
  segment_sis
};

/* bit operation */

inline static uint32_t
nth_bit(const uint8_t *key, uint32_t n)
{
  return key[n >> 3] & (0x80 >> (n & 7));
}

/* segment operation */

inline static sen_rc
load_all_segments(sen_sym *sym)
{
  sen_rc rc = sen_success;
  int seg, nkey = 0, npat = 0, nsis = 0, empty = 0;
  for (seg = 0; seg < SEN_SYM_MAX_SEGMENT; seg++) {
    switch (sym->header->segments[seg]) {
    case segment_empty :
      empty++;
      break;
    case segment_key :
      sym->keyarray[nkey++].segno = seg;
      break;
    case segment_pat :
      sym->patarray[npat++].segno = seg;
      break;
    case segment_sis :
      sym->sisarray[nsis++].segno = seg;
      break;
    default :
      rc = sen_invalid_format;
      break;
    }
  }
  return rc;
}

inline static sen_rc
segment_new(sen_sym *sym, int segtype)
{
  int seg = 0, type, n = 0;
  while ((type = sym->header->segments[seg])) {
    if (type == segtype) { n++; }
    if (++seg >= SEN_SYM_MAX_SEGMENT) { return sen_memory_exhausted; }
  }
  sym->header->segments[seg] = segtype;
  switch (segtype) {
  case segment_key :
    sym->keyarray[n].segno = seg;
    sym->header->curr_key = n * SEN_SYM_SEGMENT_SIZE;
    break;
  case segment_pat :
    sym->patarray[n].segno = seg;
    break;
  case segment_sis :
    sym->sisarray[n].segno = seg;
    break;
  }
  return sen_success;
}

/* patricia array operation */

inline static pat_node *
pat_at(sen_sym *sym, sen_id id)
{
  sen_io_seginfo *si;
  if (id > SEN_SYM_MAX_ID) { return NULL; }
  si = &sym->patarray[id >> W_OF_PAT_IN_A_SEGMENT];
  if (!si->addr) {
    if (si->segno == -1) { load_all_segments(sym); }
    if (sen_io_seg_bind(sym->io, si)) { return NULL; }
  }
  return (pat_node *)(si->addr + ((id & PAT_MASK_IN_A_SEGMENT) * sizeof(pat_node)));
}

inline static pat_node *
pat_get(sen_sym *sym, sen_id id)
{
  sen_io_seginfo *si;
  if (id > SEN_SYM_MAX_ID) { return NULL; }
  si = &sym->patarray[id >> W_OF_PAT_IN_A_SEGMENT];
  if (!si->addr) {
    while (si->segno == -1) {
      if (segment_new(sym, segment_pat)) { return NULL; }
    }
    if (sen_io_seg_bind(sym->io, si)) { return NULL; }
  }
  return (pat_node *)(si->addr + ((id & PAT_MASK_IN_A_SEGMENT) * sizeof(pat_node)));
}

inline static pat_node *
pat_node_new(sen_sym *sym, sen_id *id)
{
  uint32_t n = sym->header->curr_rec + 1;
  pat_node *res;
  if (n > SEN_SYM_MAX_ID) { return NULL; }
  if ((res = pat_get(sym, n))) {
    sym->header->curr_rec = n;
    sym->header->nrecords++;
  }
  if (id) { *id = n; }
  return res;
}

/* sis operation */

inline static sis_node *
sis_at(sen_sym *sym, sen_id id)
{
  sen_io_seginfo *si;
  if (id > SEN_SYM_MAX_ID) { return NULL; }
  si = &sym->sisarray[id >> W_OF_SIS_IN_A_SEGMENT];
  if (!si->addr) {
    if (si->segno == -1) { load_all_segments(sym); }
    if (sen_io_seg_bind(sym->io, si)) { return NULL; }
  }
  return (sis_node *)(si->addr + ((id & SIS_MASK_IN_A_SEGMENT) * sizeof(sis_node)));
}

inline static sis_node *
sis_get(sen_sym *sym, sen_id id)
{
  sen_io_seginfo *si;
  if (id == SEN_SYM_NIL || SEN_SYM_MAX_ID < id) { return NULL; }
  si = &sym->sisarray[id >> W_OF_SIS_IN_A_SEGMENT];
  if (!si->addr) {
    while (si->segno == -1) {
      if (segment_new(sym, segment_sis)) { return NULL; }
    }
    if (sen_io_seg_bind(sym->io, si)) { return NULL; }
  }
  return (sis_node *)(si->addr + ((id & SIS_MASK_IN_A_SEGMENT) * sizeof(sis_node)));
}

static void
sis_collect(sen_sym *sym, sen_set *h, sen_id id, uint32_t level)
{
  uint32_t *offset;
  sis_node *sl = sis_at(sym, id);
  if (sl) {
    sen_id sid = sl->parent;
    while (sid && sid != id) {
      sen_set_get(h, &sid, (void **) &offset);
      *offset = level;
      sis_collect(sym, h, sid, level + 1);
      if (!(sl = sis_at(sym, sid))) { break; }
      sid = sl->sibling;
    }
  }
}

/* key operation */

inline static uint8_t *
key_at(sen_sym *sym, uint32_t pos)
{
  sen_io_seginfo *si;
  if (pos > SEN_SYM_MAX_POS) { return NULL; }
  si = &sym->keyarray[pos >> W_OF_KEY_IN_A_SEGMENT];
  if (!si->addr) {
    if (si->segno == -1) { load_all_segments(sym); }
    if (sen_io_seg_bind(sym->io, si)) { return NULL; }
  }
  return (char *)(si->addr + (pos & KEY_MASK_IN_A_SEGMENT));
}

inline static uint32_t
key_put(sen_sym *sym, const uint8_t *key, int len)
{
  int lpos;
  uint32_t res = sym->header->curr_key;
  if (len >= SEN_SYM_SEGMENT_SIZE) { return 0; }
  lpos = res & KEY_MASK_IN_A_SEGMENT;
  if (lpos + len > SEN_SYM_SEGMENT_SIZE || !lpos) {
    if (segment_new(sym, segment_key)) { return 0; }
    res = sym->header->curr_key;
  }
  memcpy(key_at(sym, res), key, len);
  sym->header->curr_key += len;
  return res;
}

inline static uint8_t *
pat_node_get_key(sen_sym *sym, pat_node *n)
{
  return n->immediate ? (uint8_t *)&n->key : key_at(sym, n->key);
}

inline static sen_rc
pat_node_set_key(sen_sym *sym, pat_node *n, const uint8_t *key, int len)
{
  if (!key || !len) { return sen_invalid_argument; }
  if (len <= sizeof(uint32_t)) {
    n->immediate = 1;
    memcpy(&n->key, key, len);
  } else {
    n->immediate = 0;
    n->key = key_put(sym, key, len);
  }
  return sen_success;
}

/* delinfo operation */

inline static syn_sym_delinfo *
delinfo_search(sen_sym *sym, sen_id id)
{
  int i;
  syn_sym_delinfo *di;
  for (i = (sym->header->curr_del2) % SEN_SYM_NDELINFOS;
       i != sym->header->curr_del;
       i = (i + 1) % SEN_SYM_NDELINFOS) {
    di = &sym->header->delinfos[i];
    if (di->status != delinfo_phase1) { continue; }
    if (di->ld == id) { return di; }
    if (di->d == id) { return di; }
  }
  return NULL;
}

inline static sen_rc
delinfo_turn_2(sen_sym *sym, syn_sym_delinfo *di)
{
  sen_id d, *p = NULL;
  pat_node *ln, *dn;
  if (di->status != delinfo_phase1) { return sen_success; }
  if (!(ln = pat_at(sym, di->ld))) { return sen_invalid_argument; }
  if (!(d = di->d)) { return sen_invalid_argument; }
  if (!(dn = pat_at(sym, d))) { return sen_invalid_argument; }
  /* for check only */
  if (sen_sym_at(sym, pat_node_get_key(sym, dn))) { /* arienai */
    sen_id r;
    int c, len = sym->key_size * 8;
    pat_node *rn;
    const uint8_t *key = pat_node_get_key(sym, dn);
    printf("deleted.. %d(%s)->%d\n", d, key, sen_sym_at(sym, key));
    if (!len) { len = (strlen((char *)key) + 1) * 8; }
    rn = pat_at(sym, 0);
    c = -1;
    r = rn->r;
    while (r) {
      if (!(rn = pat_at(sym, r))) { break; }
      printf("r=%d, check=%d, ->r=%d, ->l=%d\n", r, rn->check, rn->r, rn->l);
      if (len <= rn->check) { break; }
      if (c < rn->check) {
	c = rn->check;
	r = nth_bit(key, c) ? rn->r : rn->l;
	continue;
      }
      if (!memcmp(pat_node_get_key(sym, rn), key, len >> 3)) {
	printf("found r=%d\n", r);
      }
      break;
    }
  }
  /* end of check */
  ln->deleting = 0;
  dn->deleting = 0;
  {
    sen_id r, *p0;
    pat_node *rn;
    int c, len = sym->key_size * 8;
    const uint8_t *key = pat_node_get_key(sym, dn);
    if (!key) { return sen_invalid_argument; }
    if (!len) { len = (strlen((char *)key) + 1) * 8; }
    rn = pat_at(sym, 0);
    c = -1;
    p0 = &rn->r;
    while ((r = *p0)) {
      if (r == d) {
	p = p0;
	break;
      }
      if (!(rn = pat_at(sym, r))) { return sen_invalid_format; }
      if (len <= rn->check) { return sen_invalid_format; }
      if (c >= rn->check) { return sen_invalid_format; }
      c = rn->check;
      p0 = nth_bit(key, c) ? &rn->r : &rn->l;
    }
  }
  if (p) {
    ln->check = dn->check;
    ln->r = dn->r;
    ln->l = dn->l;
    *p = di->ld;
  } else {
    /* debug */
    int j;
    sen_id dd;
    syn_sym_delinfo *ddi;
    printf("failed to find d=%d\n", d);
    for (j = (sym->header->curr_del2 + 1) % SEN_SYM_NDELINFOS;
	 j != sym->header->curr_del;
	 j = (j + 1) % SEN_SYM_NDELINFOS) {
      ddi = &sym->header->delinfos[j];
      if (ddi->status != delinfo_phase1) { continue; }
      if (!(ln = pat_at(sym, ddi->ld))) { continue; }
      if (!(dd = ddi->d)) { continue; }
      if (d == ddi->ld) {
	printf("found!!!, d(%d) become ld of (%d)\n", d, dd);
      }
    }
    /* debug */
  }
  di->status = delinfo_phase2;
  di->d = d;
  return sen_success;
}

inline static sen_rc
delinfo_turn_3(sen_sym *sym, syn_sym_delinfo *di)
{
  pat_node *dn;
  uint32_t size;
  const uint8_t *key;
  if (di->status != delinfo_phase2) { return sen_success; }
  if (!(dn = pat_at(sym, di->d))) { return sen_invalid_argument; }
  if (!(key = pat_node_get_key(sym, dn))) { return sen_invalid_argument; }
  if (sym->key_size) {
    size = sym->key_size;
  } else {
    size = dn->check = strlen((char *)key) + 1;
  }
  di->status = delinfo_empty;
  dn->r = SEN_SYM_DELETED;
  dn->l = sym->header->garbages[size];
  sym->header->garbages[size] = di->d;
  return sen_success;
}

inline static syn_sym_delinfo *
delinfo_new(sen_sym *sym)
{
  syn_sym_delinfo *res = &sym->header->delinfos[sym->header->curr_del];
  uint32_t n = (sym->header->curr_del + 1) % SEN_SYM_NDELINFOS;
  int gap = ((n + SEN_SYM_NDELINFOS - sym->header->curr_del2) % SEN_SYM_NDELINFOS)
            - (SEN_SYM_NDELINFOS / 2);
  while (gap-- > 0) {
    if (delinfo_turn_2(sym, &sym->header->delinfos[sym->header->curr_del2])) {
      printf("d2 failed: %d\n", sym->header->delinfos[sym->header->curr_del2].ld);
    }
    sym->header->curr_del2 = (sym->header->curr_del2 + 1) % SEN_SYM_NDELINFOS;
  }
  if (n == sym->header->curr_del3) {
    if (delinfo_turn_3(sym, &sym->header->delinfos[sym->header->curr_del3])) {
      printf("d3 failed: %d\n", sym->header->delinfos[sym->header->curr_del3].ld);
    }
    sym->header->curr_del3 = (sym->header->curr_del3 + 1) % SEN_SYM_NDELINFOS;
  }
  sym->header->curr_del = n;
  return res;
}

/* sym operation */

sen_sym *
sen_sym_create(const char *path, uint32_t key_size,
	       uint32_t flags, sen_encoding encoding)
{
  int i;
  sen_io *io;
  sen_sym *sym = NULL;
  pat_node *node0;
  struct sen_sym_header *header;
  io = sen_io_create(path, SEN_SYM_HEADER_SIZE, 0x400000, 8192, sen_io_auto, 8192);
  if (!io) { return NULL; }
  header = sen_io_header(io);
  memcpy(header->idstr, SEN_IDSTR, 16);
  header->flags = flags;
  header->encoding = encoding;
  header->key_size = key_size;
  header->nrecords = 0;
  header->curr_rec = 0;
  header->curr_key = -1;
  header->curr_del = 0;
  header->curr_del2 = 0;
  header->curr_del3 = 0;
  for (i = 0; i < SEN_SYM_MAX_SEGMENT; i++) {
    header->segments[i] = segment_empty;
  }
  if (!(sym = SEN_MALLOC(sizeof(sen_sym)))) {
    sen_io_close(io);
    return NULL;
  }
  sym->io = io;
  sym->header = header;
  sym->key_size = key_size;
  sym->encoding = encoding;
  sym->flags = flags;
  for (i = 0; i < SEN_SYM_MAX_SEGMENT; i++) {
    sym->keyarray[i].segno = -1;
    sym->keyarray[i].addr = NULL;
    sym->patarray[i].segno = -1;
    sym->patarray[i].addr = NULL;
    sym->sisarray[i].segno = -1;
    sym->sisarray[i].addr = NULL;
  }
  node0 = pat_get(sym, 0);
  node0->r = 0;
  node0->l = 0;
  node0->key = 0;
  return sym;
}

sen_sym *
sen_sym_open(const char *path)
{
  int i;
  sen_io *io;
  sen_sym *sym = NULL;
  struct sen_sym_header *header;
  io = sen_io_open(path, sen_io_auto, 8192);
  if (!io) { return NULL; }
  header = sen_io_header(io);
  /*
  if (memcmp(header->idstr, SEN_IDSTR, 16)) {
    puts("id not match");
    sen_io_close(io);
    return NULL;
  }
  */
  if (!(sym = SEN_MALLOC(sizeof(sen_sym)))) {
    puts("malloc failed");
    sen_io_close(io);
    return NULL;
  }
  sym->io = io;
  sym->header = header;
  sym->key_size = header->key_size;
  sym->encoding = header->encoding;
  sym->flags = header->flags;
  for (i = 0; i < SEN_SYM_MAX_SEGMENT; i++) {
    sym->keyarray[i].segno = -1;
    sym->keyarray[i].addr = NULL;
    sym->patarray[i].segno = -1;
    sym->patarray[i].addr = NULL;
    sym->sisarray[i].segno = -1;
    sym->sisarray[i].addr = NULL;
  }
  load_all_segments(sym);
  return sym;
}

sen_rc
sen_sym_close(sen_sym *sym)
{
  sen_rc rc = sen_io_close(sym->io);
  SEN_FREE(sym);
  return rc;
}

sen_rc
sen_sym_remove(const char *path)
{
  if (!path) { return sen_invalid_argument; }
  unlink(path);
  return sen_success;
}

inline static sen_id
_sen_sym_get(sen_sym *sym, const uint8_t *key, uint32_t *new)
{
  sen_id r, r0, *p0, *p1 = NULL;
  pat_node *rn, *rn0;
  int c = -1, c0 = -1, c1 = -1, len, size = sym->key_size;
  *new = 0;
  if (!key) { return SEN_SYM_NIL; }
  if (!size) { size = strlen((char *)key) + 1; }
  len = size * 8;
  if (len > SEN_SYM_MAX_KEY_LENGTH) { return SEN_SYM_NIL; }
  rn0 = pat_at(sym, 0); p0 = &rn0->r;
  // printf("key='%s', p0=%d\n", key, *p0);
  if (*p0) {
    char xor, mask;
    const char *s, *d;
    while ((r0 = *p0)) {
      if (!(rn0 = pat_at(sym, r0))) { return 0; }
      // printf("%d[%d,%d] key='%s' c0=%d c=%d len=%d\n", r0, rn0->r, rn0->l, pat_node_get_key(sym, rn0), c0, rn0->check, len);
      if (c0 < rn0->check && rn0->check < len) {
	c1 = c0; c0 = rn0->check;
	p1 = p0; p0 = nth_bit(key, c0) ? &rn0->r : &rn0->l;
      } else {
	if (rn0->check < len && !memcmp(pat_node_get_key(sym, rn0), key, size)) {
	  return r0;
	}
	break;
      }
    }
    for (c = 0, s = pat_node_get_key(sym, rn0), d = key; *s == *d; c += 8, s++, d++);
    for (xor = *s ^ *d, mask = 0x80; !(xor & mask); mask >>= 1, c++);
    // printf("'%s', '%s', check=%d, c0=%d, c1=%d\n", key, pat_node_get_key(sym, rn0), c, c0, c1);
    if (c < c0) {
      if (c > c1) {
	p0 = p1;
      } else {
	rn0 = pat_at(sym, 0); p0 = &rn0->r;
	while ((r0 = *p0)) {
	  if (!(rn0 = pat_at(sym, r0))) { return 0; }
	  if (c < rn0->check) { break; }
	  p0 = nth_bit(key, rn0->check) ? &rn0->r : &rn0->l;
	}
	// printf("ins %d[%d,%d]%d\n", r0, rn0->r, rn0->l, rn0->check);
      }
    }
    if (c >= len || c > SEN_SYM_MAX_KEY_LENGTH) { return 0; }
  } else {
    c = (len - 1 > SEN_SYM_MAX_KEY_LENGTH) ? SEN_SYM_MAX_KEY_LENGTH : len - 1;
  }
  if (sym->header->garbages[size]) {
    r = sym->header->garbages[size];
    sym->header->nrecords++;
    rn = pat_at(sym, r);
    sym->header->garbages[size] = rn->l;
    memcpy(pat_node_get_key(sym, rn), key, size);
  } else {
    if (!(rn = pat_node_new(sym, &r))) { return 0; }
    pat_node_set_key(sym, rn, key, size);
  }
  rn->check = c;
  rn->deleting = 0;
  if (nth_bit(key, c)) {
    rn->r = r;
    rn->l = *p0;
  } else {
    rn->r = *p0;
    rn->l = r;
  }

  // printf("new=%d[%d,%d](%d)\n", r, rn->r, rn->l, rn->check);

  *p0 = r;
  *new = 1;
  return r;
}


inline static int
chop(sen_sym *sym, const char **key)
{
  size_t len = sen_str_charlen(*key, sym->encoding);
  if (len) {
    *key += len;
    return **key;
  } else {
    return 0;
  }
}

sen_id
sen_sym_get(sen_sym *sym, const uint8_t *key)
{
  uint32_t new;
  sen_id r0;
  r0 = _sen_sym_get(sym, key, &new);
  if ((sym->flags & SEN_SYM_WITH_SIS) && (*key & 0x80)) { // todo: refine!!
    sis_node *sl, *sr;
    sen_id l = r0, r;
    if (new && (sl = sis_get(sym, l))) {
      const char *sis = key;
      sl->parent = l;
      sl->sibling = 0;
      while (chop(sym, &sis)) {
	r = _sen_sym_get(sym, sis, &new);
	if (!(sr = sis_get(sym, r))) { break; }
	if (new) {
	  sl->sibling = r;
	  sr->parent = l;
	  sr->sibling = 0;
	} else {
	  sl->sibling = sr->parent;
	  sr->parent = l;
	  break;
	}
	l = r;
	sl = sr;
      }
    }
  }
  return r0;
}

sen_id
sen_sym_at(sen_sym *sym, const uint8_t *key)
{
  sen_id r;
  pat_node *rn;
  int c = -1, len = sym->key_size * 8;
  if (!key) { return SEN_SYM_NIL; }
  if (!len) { len = (strlen((char *)key) + 1) * 8; }
  for (r = pat_at(sym, 0)->r; r; r = nth_bit(key, c) ? rn->r : rn->l) {
    if (!(rn = pat_at(sym, r))) { break; /* corrupt? */ }
    if (len <= rn->check) { break; }
    if (rn->check <= c) {
      if (!memcmp(pat_node_get_key(sym, rn), key, len >> 3)) { return r; }
      break;
    }
    c = rn->check;
  }
  return SEN_SYM_NIL;
}

static void
get_tc(sen_sym *sym, sen_set *h, pat_node *rn, int len)
{
  sen_id id;
  pat_node *node;
  id = rn->r;
  if (id && (node = pat_at(sym, id))) {
    if (node->check > rn->check) {
      sen_set_get(h, &id, NULL);
      get_tc(sym, h, node, len);
    } else if (node->check < len) {
      sen_set_get(h, &id, NULL);
    }
  }
  id = rn->l;
  if (id && (node = pat_at(sym, id))) {
    if (node->check > rn->check) {
      sen_set_get(h, &id, NULL);
      get_tc(sym, h, node, len);
    } else if (node->check < len) {
      sen_set_get(h, &id, NULL);
    }
  }
}

sen_set *
sen_sym_prefix_search(sen_sym *sym, const uint8_t *key)
{
  int c, len;
  sen_id r;
  pat_node *rn;
  if (!key || sym->key_size) { return NULL; } /* only for string */
  len = strlen((char *)key) * 8;
  rn = pat_at(sym, 0);
  c = -1;
  r = rn->r;
  while (r) {
    if (!(rn = pat_at(sym, r))) { return NULL; }
    if (c < rn->check && len > rn->check) {
      c = rn->check;
      r = nth_bit(key, c) ? rn->r : rn->l;
      continue;
    }
    if (!memcmp(pat_node_get_key(sym, rn), key, len >> 3)) {
      sen_set *h = sen_set_open(sizeof(sen_id), 0, 0);
      sen_set_get(h, &r, NULL);
      if (rn->check >= len) { get_tc(sym, h, rn, len); }
      return h;
    }
    break;
  }
  return NULL;
}

sen_set *
sen_sym_suffix_search(sen_sym *sym, const uint8_t *key)
{
  sen_id r;
  uint32_t *offset;
  if (!key || sym->key_size) { return NULL; }
  if ((r = sen_sym_at(sym, key))) {
    sen_set *h = sen_set_open(sizeof(sen_id), sizeof(uint32_t), 0);
    sen_set_get(h, &r, (void **) &offset);
    *offset = 0;
    if (sym->flags & SEN_SYM_WITH_SIS) { sis_collect(sym, h, r, 1); }
    return h;
  }
  return NULL;
}

sen_id
sen_sym_common_prefix_search(sen_sym *sym, const uint8_t *key)
{
  pat_node *rn, *rnl;
  sen_id r, rl, r2 = SEN_SYM_NIL;
  int c = -1, cl, len = 0, n, len2 = 8, nth_bit;
  if (!key) { return SEN_SYM_NIL; }
  if (sym->key_size) { return SEN_SYM_NIL; /* only for string */ }
  // len = (strlen((char *)key) + 1) * 8;
  for (r = pat_at(sym, 0)->r; r;) {
    if (!(rn = pat_at(sym, r))) { break; /* corrupt? */ }
    // if (len <= rn->check) { break; }
    if (rn->check <= c) {
      uint8_t *p = pat_node_get_key(sym, rn);
      if (!memcmp(p, key, strlen(p))) { return r; }
      break;
    }
    c = rn->check;
    n = c >> 3;
    while (len < n) { if (!key[len++]) { goto exit; } }
    nth_bit = key[n] & (0x80 >> (c & 7));
    if (nth_bit) {
      if (len2 <= c) {
	len2 = (n + 1) * 8;
	for (cl = c, rl = rn->l; rl; cl = rnl->check, rl = rnl->l) {
	  if (!(rnl = pat_at(sym, rl))) { break; /* corrupt? */ }
	  if (len2 <= rnl->check) { break; }
	  if (rnl->check <= cl) {
	    uint8_t *p = pat_node_get_key(sym, rnl);
	    if (!p[n] && !memcmp(p, key, n)) { r2 = rl; }
	    break;
	  }
	}
      }
      r = rn->r;
    } else {
      r = rn->l;
    }
  }
exit :
  return r2;
}

sen_rc
sen_sym_del(sen_sym *sym, const uint8_t *key)
{
  syn_sym_delinfo *di;
  uint8_t direction;
  pat_node *rn, *rn0 = NULL;
  int c, len = sym->key_size * 8;
  sen_id r, otherside, *p, *p0 = NULL;
  if (!key) { return sen_invalid_argument; }
  if (!len) { len = (strlen((char *)key) + 1) * 8; }
  di = delinfo_new(sym); /* must be called before find rn */
  rn = pat_at(sym, 0);
  c = -1;
  p = &rn->r;
  for (;;) {
    if (!(r = *p)) { return sen_invalid_argument; }
    if (!(rn = pat_at(sym, r))) { return sen_invalid_format; }
    if (len <= rn->check) { return sen_invalid_argument; }
    if (c >= rn->check) {
      if (!memcmp(pat_node_get_key(sym, rn), key, len >> 3)) {
	break; /* found */
      } else { 	return sen_invalid_argument; }
    }
    c = rn->check;
    rn0 = rn;
    p0 = p; p = nth_bit(key, c) ? &rn->r : &rn->l;
  }
  direction = (rn0->r == r);
  otherside = direction ? rn0->l : rn0->r;
  if (rn == rn0) {
    di->status = delinfo_phase2;
    di->d = r;
    *p0 = otherside;
  } else {
    syn_sym_delinfo *ldi = NULL, *ddi = NULL;
    if (rn->deleting) { ldi = delinfo_search(sym, r); }
    if (rn0->deleting) { ddi = delinfo_search(sym, *p0); }
    if (ldi) {
      rn->deleting = 0;
      di->status = delinfo_phase2;
      if (ddi) {
	rn0->deleting = 0;
	ddi->status = delinfo_phase2;
	if (ddi == ldi) {
	  di->d = ddi->ld;
	} else {
	  ldi->ld = ddi->ld;
	  di->d = r;
	}
      } else {
	rn0->deleting = 1;
	ldi->ld = *p0;
	di->d = r;
      }
    } else {
      rn->deleting = 1;
      if (ddi) {
	rn0->deleting = 0;
	ddi->d = r;
	di->status = delinfo_phase2;
	di->d = *p0;
      } else {
	rn0->deleting = 1;
	di->status = delinfo_phase1;
	di->ld = *p0;
	di->d = r;
      }
    }
    if (*p0 == otherside) {
      rn0->check = 0;
    } else {
      *p0 = otherside;
    }
  }
  sym->header->nrecords--;
  return sen_success;
}

uint32_t
sen_sym_size(sen_sym *sym)
{
  return sym->header->nrecords;
}

const uint8_t *
sen_sym_key(sen_sym *sym, sen_id id)
{
  pat_node *node = pat_at(sym, id);
  if (!node) { return NULL; }
  return pat_node_get_key(sym, node);
}

uint32_t
sen_sym_pocket_get(sen_sym *sym, sen_id id)
{
  pat_node *node = pat_at(sym, id);
  if (!node) { return -1; }
  return node->pocket;
}

sen_rc
sen_sym_pocket_set(sen_sym *sym, sen_id id, uint32_t value)
{
  pat_node *node = pat_at(sym, id);
  if (!node) { return sen_invalid_argument; }
  node->pocket = value;
  return sen_success;
}

sen_rc
sen_sym_info(sen_sym *sym, int *key_size,
		    unsigned *flags, sen_encoding *encoding, unsigned *nrecords)
{
  if (!sym) { return sen_invalid_argument; }
  if (key_size) { *key_size = sym->key_size; }
  if (flags) { *flags = sym->flags; }
  if (encoding) { *encoding = sym->encoding; }
  if (nrecords) { *nrecords = sym->header->nrecords; }
  return sen_success;
}

#ifndef NO_DEBUG
void
sen_sym_dump(sen_sym *sym, int to)
{
  sen_id i;
  pat_node *rn;
  if (to < 0) { to = sym->header->curr_rec; }
  for (i = 0; i <= to; i++) {
    rn = pat_at(sym, i);
    printf("%d(%p):[%d,%d](%d)\"%s\"\n",
            i, rn, rn->r, rn->l, rn->check, pat_node_get_key(sym, rn));
  }
}

void
sen_sym_inspect(sen_sym *sym)
{
  int i;
  printf("idstr=%s\nkey_size=%d\nnrecords=%d\ncurr_rec=%d\ncurr_key=%d\ncurr_del=%d\ncurr_del2=%d\ncurr_del3=%d\n",
	 sym->header->idstr,
	 sym->header->key_size,
	 sym->header->nrecords,
	 sym->header->curr_rec,
	 sym->header->curr_key,
	 sym->header->curr_del,
	 sym->header->curr_del2,
	 sym->header->curr_del3);
  {
    int se = 0, ss = 0, sa = 0, si = 0;
    for (i = 0; i < SEN_SYM_MAX_SEGMENT; i++) {
      switch (sym->header->segments[i]) {
      case segment_empty :
	se++;
	break;
      case segment_key :
	ss++;
	break;
      case segment_pat :
	sa++;
	break;
      default :
	si++;
      }
    }
    printf("seg_empty=%d, seg_string=%d, seg_array=%d, seg_invalid=%d\n", se, ss, sa, si);
  }
  {
    int n = 0, s0 = 0, s1= 0, s2 = 0, s3 = 0;
    printf("del2=%d, del=%d\n", sym->header->curr_del2, sym->header->curr_del);
    for (i = sym->header->curr_del2;
	 i != sym->header->curr_del;
	 i = (i + 1) % SEN_SYM_NDELINFOS) {
      n++;
      switch (sym->header->delinfos[i].status) {
      case 0 : s0++; break;
      case 1 : s1++; break;
      case 2 : s2++; break;
      case 3 : s3++; break;
      }
    }
    printf("del2->del: n=%d, s0=%d, s1=%d, s2=%d, s3=%d\n", n, s0, s1, s2, s3);
  }
  {
    int n = 0, s0 = 0, s1= 0, s2 = 0, s3 = 0;
    for (i = sym->header->curr_del3;
	 i != sym->header->curr_del2;
	 i = (i + 1) % SEN_SYM_NDELINFOS) {
      n++;
      switch (sym->header->delinfos[i].status) {
      case 0 : s0++; break;
      case 1 : s1++; break;
      case 2 : s2++; break;
      case 3 : s3++; break;
      }
    }
    printf("del3->del2: n=%d, s0=%d, s1=%d, s2=%d, s3=%d\n", n, s0, s1, s2, s3);
  }
  {
    int n = 0, s0 = 0, s1= 0, s2 = 0, s3 = 0;
    for (i = sym->header->curr_del;
	 i != sym->header->curr_del3;
	 i = (i + 1) % SEN_SYM_NDELINFOS) {
      n++;
      switch (sym->header->delinfos[i].status) {
      case 0 : s0++; break;
      case 1 : s1++; break;
      case 2 : s2++; break;
      case 3 : s3++; break;
      }
    }
    printf("del->del3: n=%d, s0=%d, s1=%d, s2=%d, s3=%d\n", n, s0, s1, s2, s3);
  }
  {
    int n = 0, nn = 0, ndel = 0, nding = 0;
    pat_node *tn;
    for (i = 1; i <= sym->header->curr_rec; i++) {
      tn = pat_at(sym, i);
      if (tn) {
	n++;
	if (tn->r == SEN_SYM_DELETED) { ndel++; }
	if (tn->deleting) { nding++; }
      } else {
	nn++;
      }
    }
    printf("array: n=%d, nil=%d, del=%d, deleting=%d\n", n, nn, ndel, nding);
  }
  {
    int i;
    for (i = 0; i < SEN_SYM_MAX_SEGMENT; i++) {
      if (sym->keyarray[i].segno != -1) { printf("keyarray[%d]=%x\n", i, sym->keyarray[i].segno); }
    }
    for (i = 0; i < SEN_SYM_MAX_SEGMENT; i++) {
      if (sym->patarray[i].segno != -1) { printf("patarray[%d]=%x\n", i, sym->patarray[i].segno); }
    }
    for (i = 0; i < SEN_SYM_MAX_SEGMENT; i++) {
      if (sym->sisarray[i].segno != -1) { printf("sisarray[%d]=%x\n", i, sym->sisarray[i].segno); }
    }
  }
}

static void
show_sid(sen_sym *sym, sen_id id, int level)
{
  sis_node *sl = sis_at(sym, id);
  if (sl) {
    sen_id sid = sl->parent;
    while (sid && sid != id) {
      int i = level;
      while(i--) { putchar(' '); }
      puts(sen_sym_key(sym, sid));
      show_sid(sym, sid, level + 1);
      if (!(sl = sis_at(sym, sid))) { break; }
      sid = sl->sibling;
    }
  }
}

void
sen_sym_inspect_term(sen_sym *sym, sen_id id)
{
  puts(sen_sym_key(sym, id));
  if (sym->flags & SEN_SYM_WITH_SIS) {
    show_sid(sym, id, 1);
  }  
}
#endif /* NO_DEBUG */
